import numpy as np


def sum_squares(a, b):
    """
    Return the sum of squares of the difference between a and b, aka MSE
    """
    return np.sum([(a[i] - b[i])**2 for i in range(len(a))])


def euclid_d(x, y):
    """
    1D euclidean distance
    """
    return np.sqrt((x-y)**2)


def man_d(x, y):
    """
    1D manhattan distance
    """
    return np.abs((x-y))


def classif_d(x, y):
    """ 
    Function adapted to classification problems
    """
    return np.array(0 if x == y else 1)


def rmse(pred, ground_truth):
    import numpy as np
    return np.sqrt(sum_squares(pred, ground_truth)/len(ground_truth))


def accuracy(pred, ground_truth):
    import numpy as np
    return np.mean([a == b for a, b in zip(pred, ground_truth)])


def rbf_k(D, sigma=1):
    return np.exp(-(D**2)/sigma)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 26 18:26:36 2020

@author: ljia
"""
# from gklearn.utils import Dataset

class PreimageGenerator(object):
	
	def __init__(self, dataset=None):
		# arguments to set.
		self._dataset = None if dataset is None else dataset
		self._kernel_options = {}
		self._graph_kernel = None
		self._verbose = 2

	@property
	def dataset(self):
		return self._dataset

	@dataset.setter
	def dataset(self, value):
		self._dataset = value		
		
	
	@property
	def kernel_options(self):
		return self._kernel_options
	
	@kernel_options.setter
	def kernel_options(self, value):
		self._kernel_options = value


	@property
	def graph_kernel(self):
		return self._graph_kernel
		
		
	@property
	def verbose(self):
		return self._verbose

	@verbose.setter
	def verbose(self, value):
		self._verbose = value




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 27 18:59:28 2020

@author: ljia
"""
import os
import datetime
import time
import sys
from tqdm import tqdm
from multiprocessing import Pool, Array
from functools import partial
import numpy as np
from matplotlib import pyplot as plt
from sklearn.model_selection import KFold, train_test_split, ParameterGrid
from sklearn.kernel_ridge import KernelRidge
from sklearn.svm import SVC
from sklearn.metrics import accuracy_score, mean_squared_error


class NestedCV(object):
	"""Perform model selection, fitting and testing for precomputed kernels
	using nested CV. Print out neccessary data during the process then finally
	the results.

	Parameters
	----------
	datafile : string
		Path of dataset file.
	estimator : function
		kernel function used to estimate. This function needs to return a gram matrix.
	param_grid_precomputed : dictionary
		Dictionary with names (string) of parameters used to calculate gram
		matrices as keys and lists of parameter settings to try as values. This
		enables searching over any sequence of parameter settings. Params with
		length 1 will be omitted.
	param_grid : dictionary
		Dictionary with names (string) of parameters used as penelties as keys
		and lists of parameter settings to try as values. This enables
		searching over any sequence of parameter settings. Params with length 1
		will be omitted.
	model_type : string
		Type of the problem, can be 'regression' or 'classification'.
	NUM_TRIALS : integer
		Number of random trials of the outer CV loop. The default is 30.
	datafile_y : string
		Path of file storing y data. This parameter is optional depending on
		the given dataset file.
	extra_params : dict
		Extra parameters for loading dataset. See function gklearn.utils.
		graphfiles.loadDataset for detail.
	ds_name : string
		Name of the dataset.
	n_jobs : int
		Number of jobs for parallelization.
	read_gm_from_file : boolean
		Whether gram matrices are loaded from a file.

	Examples
	--------
	>>> import numpy as np
	>>> from gklearn.utils.model_selection_precomputed import model_selection_for_precomputed_kernel
	>>> from gklearn.kernels.untilHPathKernel import untilhpathkernel
	>>>
	>>> datafile = '../datasets/MUTAG/MUTAG_A.txt'
	>>> estimator = untilhpathkernel
	>>> param_grid_precomputed = {’depth’:  np.linspace(1, 10, 10), ’k_func’:
			[’MinMax’, ’tanimoto’], ’compute_method’:  [’trie’]}
	>>> # ’C’ for classification problems and ’alpha’ for regression problems.
	>>> param_grid = [{’C’: np.logspace(-10, 10, num=41, base=10)}, {’alpha’:
			np.logspace(-10, 10, num=41, base=10)}]
	>>>
	>>> model_selection_for_precomputed_kernel(datafile, estimator,
			param_grid_precomputed, param_grid[0], 'classification', ds_name=’MUTAG’)
	"""
	def __init__(self, dataset, estimator, param_grid_precomputed=None, param_grid=None, model_type=None, num_trials=30, output_dir=None, n_jobs=1, save_gms=True, save_gm_figs=False, logging=True, verbose=True,  **kwargs):
		tqdm.monitor_interval = 0
		self._ds = dataset
		self._estimator = estimator
		self._num_trials = num_trials
		self._n_jobs = n_jobs
		self._save_gms = save_gms
		self._save_gm_figs = save_gm_figs
		self._logging = logging
		self._verbose = verbose
		self._kwargs = kwargs

		# Set dataset name.
		if self._ds._ds_name is None:
			self._ds_name = 'ds-unknown'
		else:
			self._ds_name = self._ds._ds_name

		# The output directory.
		if output_dir is None:
			self._output_dir = os.path.join('outputs/', estimator.__name__)
		else:
			self._output_dir = output_dir
		os.makedirs(self._output_dir, exist_ok=True)

		# Setup the model type.
		if model_type is None:
			self._model_type = dataset._task_type
		else:
			self._model_type = model_type.lower()
		if self._model_type != 'regression' and self._model_type != 'classification':
			raise Exception('The model type is incorrect! Please choose from regression or classification.')

		# @todo: Set param_grid_precomputed and param_grid.
		self._param_grid_precomputed = param_grid_precomputed
		self._param_grid = param_grid

		if self._verbose:
			print()
			print('--- This is a %s problem ---' % self._model_type)
		# A string to save all the results.
		if self._logging:
			self._str_fw = '###################### log time: ' + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + '. ######################\n\n'
			self._str_fw += '# This file contains results of ' + self._estimator.__name__ + ' on dataset ' + self._ds_name + ',\n# including gram matrices, serial numbers for gram matrix figures and performance.\n\n'
			self._str_fw += 'This is a %s problem.\n' % self._model_type

		self.run()


	def run(self):
		self.fit()
		self.compute_gram_matrices()
		if len(self._gram_matrices) == 0:
			if self._verbose:
				print('All gram matrices are ignored, no results obtained.')
			if self._logging:
				self._str_fw += '\nAll gram matrices are ignored, no results obtained.\n\n'
		else:
			self.do_cv()

		# print out results as table.
		if self._logging:
			self._str_fw += self.printResultsInTable(self._param_list, self._param_list_pre_revised, self._average_val_scores, self._std_val_scores, self._average_perf_scores, self._std_perf_scores, self._average_train_scores, self._std_train_scores, self._gram_matrix_time, self._model_type, self._verbose)

			# open file to save all results for this dataset.
			if not os.path.exists(self._output_dir + '/' + self._ds_name + '.output.txt'):
				with open(self._output_dir + '/' + self._ds_name + '.output.txt', 'w') as f:
					f.write(self._str_fw)
			else:
				with open(self._output_dir + '/' + self._ds_name + '.output.txt', 'r+') as f:
					content = f.read()
					f.seek(0, 0)
					f.write(self._str_fw + '\n\n\n' + content)

		return self._final_performance, self._final_confidence


	def fit(self):
		return


	def compute_gram_matrices(self):
		"""Compute all gram matrices.

		Returns
		-------
		None.

		"""
		# Grid of parameters with a discrete number of values for each.
		self._param_list_precomputed = list(ParameterGrid(self._param_grid_precomputed))
		self._param_list = list(ParameterGrid(self._param_grid))

		self._gram_matrices = [
		]  # a list to store gram matrices for all param_grid_precomputed
		self._gram_matrix_time = [
		]  # a list to store time to calculate gram matrices
		self._param_list_pre_revised = [
		]  # list to store param grids precomputed ignoring the useless ones

		if self._verbose:
			print()
			print('\n1. Computing gram matrices. This could take a while...')
		if self._logging:
			self._str_fw += '\nI. Gram matrices.\n\n'
		self._tts = time.time()  # start training time
		nb_gm_ignore = 0  # the number of gram matrices those should not be considered, as they may contain elements that are not numbers (NaN)
		for idx, params_out in enumerate(self._param_list_precomputed):
			y = self._ds.targets[:]
			params_out['n_jobs'] = self._n_jobs
			params_out['verbose'] = self._verbose
#			print(dataset)
#			import networkx as nx
#			nx.draw_networkx(dataset[1])
#			plt.show()
			rtn_data = self._estimator(self._ds.graphs[:], **params_out) # @todo: Attention! this will not copy the graphs.
			Kmatrix = rtn_data[0]
			current_run_time = rtn_data[1]
			# for some kernels, some graphs in datasets may not meet the
			# kernels' requirements for graph structure. These graphs are trimmed.
			if len(rtn_data) == 3:
				idx_trim = rtn_data[2]  # the index of trimmed graph list
				y = [y[idxt] for idxt in idx_trim] # trim y accordingly
#			Kmatrix = np.random.rand(2250, 2250)
#			current_run_time = 0.1

			# remove graphs whose kernels with themselves are zeros
			# @todo: y not changed accordingly?
			Kmatrix_diag = Kmatrix.diagonal().copy()
			nb_g_ignore = 0
			for idxk, diag in enumerate(Kmatrix_diag):
				if diag == 0:
					Kmatrix = np.delete(Kmatrix, (idxk - nb_g_ignore), axis=0)
					Kmatrix = np.delete(Kmatrix, (idxk - nb_g_ignore), axis=1)
					nb_g_ignore += 1

			# normalization
			# @todo: works only for undirected graph?
			Kmatrix_diag = Kmatrix.diagonal().copy()
			for i in range(len(Kmatrix)):
				for j in range(i, len(Kmatrix)):
					Kmatrix[i][j] /= np.sqrt(Kmatrix_diag[i] * Kmatrix_diag[j])
					Kmatrix[j][i] = Kmatrix[i][j]
			if self._verbose:
				print()

			if params_out == {}:
				if self._verbose:
					print('the gram matrix is: ')
				if self._logging:
					self._str_fw += 'the gram matrix is:\n\n'
			else:
				if self._verbose:
					print('the gram matrix with parameters', params_out, 'is: \n\n')
				if self._logging:
					self._str_fw += 'the gram matrix with parameters %s is:\n\n' % params_out

			if len(Kmatrix) < 2:
				nb_gm_ignore += 1
				if self._verbose:
					print('ignored, as at most only one of all its diagonal value is non-zero.')
				if self._logging:
					self._str_fw += 'ignored, as at most only one of all its diagonal value is non-zero.\n\n'
			else:
				if np.isnan(Kmatrix).any(
				):  # if the matrix contains elements that are not numbers
					nb_gm_ignore += 1
					if self._verbose:
						print('ignored, as it contains elements that are not numbers.')
					if self._logging:
						self._str_fw += 'ignored, as it contains elements that are not numbers.\n\n'
				else:
#					print(Kmatrix)
					if self._logging:
						self._str_fw += np.array2string(
							Kmatrix,
							separator=',') + '\n\n'
#							separator=',',
#							threshold=np.inf,
#							floatmode='unique') + '\n\n'

					# Draw and save Gram matrix figures.
					if self._save_gm_figs:
						fig_file_name = self._output_dir + '/GM[ds]' + self._ds_name
						if params_out != {}:
							fig_file_name += '[params]' + str(idx)
						plt.imshow(Kmatrix)
						plt.colorbar()
						plt.savefig(fig_file_name + '.eps', format='eps', dpi=300)
	#					plt.show()
						plt.clf()

					self._gram_matrices.append(Kmatrix)
					self._gram_matrix_time.append(current_run_time)
					self._param_list_pre_revised.append(params_out)

					if nb_g_ignore > 0:
						if self._verbose:
							print(', where %d graphs are ignored as their graph kernels with themselves are zeros.' % nb_g_ignore)
						if self._logging:
							self._str_fw += ', where %d graphs are ignored as their graph kernels with themselves are zeros.' % nb_g_ignore

		if self._verbose:
			print()
			print('{} gram matrices are calculated, {} of which are ignored.'.format(len(self._param_list_precomputed), nb_gm_ignore))
		if self._logging:
			self._str_fw += '{} gram matrices are calculated, {} of which are ignored.\n\n'.format(len(self._param_list_precomputed), nb_gm_ignore)
			self._str_fw += 'serial numbers of gram matrix figures and their corresponding parameters settings:\n\n'
			self._str_fw += ''.join(['{}: {}\n'.format(idx, params_out) for idx, params_out in enumerate(self._param_list_precomputed)])


	def do_cv(self):
# save gram matrices to file.
#			np.savez(output_dir + '/' + ds_name + '.gm',
#					 gms=gram_matrices, params=param_list_pre_revised, y=y,
#					 gmtime=gram_matrix_time)
		if self._verbose:
			print('2. Fitting and predicting using nested cross validation. This could really take a while...')

		# ---- use pool.imap_unordered to parallel and track progress. ----
#			train_pref = []
#			val_pref = []
#			test_pref = []
#			def func_assign(result, var_to_assign):
#				for idx, itm in enumerate(var_to_assign):
#					itm.append(result[idx])
#			trial_do_partial = partial(trial_do, param_list_pre_revised, param_list, y, model_type)
#
#			parallel_me(trial_do_partial, range(NUM_TRIALS), func_assign,
#						[train_pref, val_pref, test_pref], glbv=gram_matrices,
#						method='imap_unordered', n_jobs=n_jobs, chunksize=1,
#						itr_desc='cross validation')

		def init_worker(gms_toshare):
			global G_gms
			G_gms = gms_toshare

#			gram_matrices = np.array(gram_matrices)
#			gms_shape = gram_matrices.shape
#			gms_array = Array('d', np.reshape(gram_matrices.copy(), -1, order='C'))
#			pool = Pool(processes=n_jobs, initializer=init_worker, initargs=(gms_array, gms_shape))
		pool = Pool(processes=self._n_jobs, initializer=init_worker, initargs=(self._gram_matrices,))
		trial_do_partial = partial(self._parallel_trial_do, self._param_list_pre_revised, self._param_list, self._ds.targets[:], self._model_type) # @todo: maybe self._ds.targets[:] should be y.
		train_pref = []
		val_pref = []
		test_pref = []
#			if NUM_TRIALS < 1000 * n_jobs:
#				chunksize = int(NUM_TRIALS / n_jobs) + 1
#			else:
#				chunksize = 1000
		chunksize = 1
		if self._verbose:
			iterator = tqdm(pool.imap_unordered(trial_do_partial, range(self._num_trials), chunksize), desc='cross validation', file=sys.stdout)
		else:
			iterator = pool.imap_unordered(trial_do_partial, range(self._num_trials), chunksize)
		for o1, o2, o3 in iterator:
			train_pref.append(o1)
			val_pref.append(o2)
			test_pref.append(o3)
		pool.close()
		pool.join()

#			# ---- use pool.map to parallel. ----
#			pool =  Pool(n_jobs)
#			trial_do_partial = partial(trial_do, param_list_pre_revised, param_list, gram_matrices, y[0:250], model_type)
#			result_perf = pool.map(trial_do_partial, range(NUM_TRIALS))
#			train_pref = [item[0] for item in result_perf]
#			val_pref = [item[1] for item in result_perf]
#			test_pref = [item[2] for item in result_perf]

#			# ---- direct running, normally use a single CPU core. ----
#			train_pref = []
#			val_pref = []
#			test_pref = []
#			for i in tqdm(range(NUM_TRIALS), desc='cross validation', file=sys.stdout):
#				o1, o2, o3 = trial_do(param_list_pre_revised, param_list, gram_matrices, y, model_type, i)
#				train_pref.append(o1)
#				val_pref.append(o2)
#				test_pref.append(o3)
#			print()

		if self._verbose:
			print()
			print('3. Getting final performance...')
		if self._logging:
			self._str_fw += '\nII. Performance.\n\n'

		# averages and confidences of performances on outer trials for each combination of parameters
		self._average_train_scores = np.mean(train_pref, axis=0)
#			print('val_pref: ', val_pref[0][0])
		self._average_val_scores = np.mean(val_pref, axis=0)
#			print('test_pref: ', test_pref[0][0])
		self._average_perf_scores = np.mean(test_pref, axis=0)
		# sample std is used here
		self._std_train_scores = np.std(train_pref, axis=0, ddof=1)
		self._std_val_scores = np.std(val_pref, axis=0, ddof=1)
		self._std_perf_scores = np.std(test_pref, axis=0, ddof=1)

		if self._model_type == 'regression':
			best_val_perf = np.amin(self._average_val_scores)
		else:
			best_val_perf = np.amax(self._average_val_scores)
#			print('average_val_scores: ', self._average_val_scores)
#			print('best_val_perf: ', best_val_perf)
#			print()
		best_params_index = np.where(self._average_val_scores == best_val_perf)
		# find smallest val std with best val perf.
		best_val_stds = [
			self._std_val_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]
		min_val_std = np.amin(best_val_stds)
		best_params_index = np.where(self._std_val_scores == min_val_std)
		best_params_out = [self._param_list_pre_revised[i] for i in best_params_index[0]]
		best_params_in = [self._param_list[i] for i in best_params_index[1]]

		if self._verbose:
			print('best_params_out: ', best_params_out)
			print('best_params_in: ', best_params_in)
			print()
			print('best_val_perf: ', best_val_perf)
			print('best_val_std: ', min_val_std)
		if self._logging:
			self._str_fw += 'best settings of hyper-params to build gram matrix: %s\n' % best_params_out
			self._str_fw += 'best settings of other hyper-params: %s\n\n' % best_params_in
			self._str_fw += 'best_val_perf: %s\n' % best_val_perf
			self._str_fw += 'best_val_std: %s\n' % min_val_std

#			print(best_params_index)
#			print(best_params_index[0])
#			print(self._average_perf_scores)
		self._final_performance = [
			self._average_perf_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]
		self._final_confidence = [
			self._std_perf_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]

		if self._verbose:
			print('final_performance: ', self._final_performance)
			print('final_confidence: ', self._final_confidence)
		if self._logging:
			self._str_fw += 'final_performance: %s\n' % self._final_performance
			self._str_fw += 'final_confidence: %s\n' % self._final_confidence

		train_performance = [
			self._average_train_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]
		train_std = [
			self._std_train_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]

		if self._verbose:
			print('train_performance: %s' % train_performance)
			print('train_std: ', train_std)
		if self._logging:
			self._str_fw += 'train_performance: %s\n' % train_performance
			self._str_fw += 'train_std: %s\n\n' % train_std

		if self._verbose:
			print()

		tt_total = time.time() - self._tts  # training time for all hyper-parameters
		average_gram_matrix_time = np.mean(self._gram_matrix_time)
		std_gram_matrix_time = np.std(self._gram_matrix_time, ddof=1) if len(self._gram_matrix_time) > 1 else 0
		best_gram_matrix_time = [self._gram_matrix_time[i] for i in best_params_index[0]]
		ave_bgmt = np.mean(best_gram_matrix_time)
		std_bgmt = np.std(best_gram_matrix_time, ddof=1) if len(best_gram_matrix_time) > 1 else 0

		if self._verbose:
			print('time to calculate gram matrix with different hyper-params: {:.2f}±{:.2f}s'
				  .format(average_gram_matrix_time, std_gram_matrix_time))
			print('time to calculate best gram matrix: {:.2f}±{:.2f}s'.format(
					ave_bgmt, std_bgmt))
			print('total training time with all hyper-param choices: {:.2f}s'.format(
					tt_total))
		if self._logging:
			self._str_fw += 'time to calculate gram matrix with different hyper-params: {:.2f}±{:.2f}s\n'.format(average_gram_matrix_time, std_gram_matrix_time)
			self._str_fw += 'time to calculate best gram matrix: {:.2f}±{:.2f}s\n'.format(ave_bgmt, std_bgmt)
			self._str_fw += 'total training time with all hyper-param choices: {:.2f}s\n\n'.format(tt_total)

		# # save results to file
		# np.savetxt(results_name_pre + 'average_train_scores.dt',
		#			average_train_scores)
		# np.savetxt(results_name_pre + 'average_val_scores', self._average_val_scores)
		# np.savetxt(results_name_pre + 'average_perf_scores.dt',
		#			average_perf_scores)
		# np.savetxt(results_name_pre + 'std_train_scores.dt', self._std_train_scores)
		# np.savetxt(results_name_pre + 'std_val_scores.dt', self._std_val_scores)
		# np.savetxt(results_name_pre + 'std_perf_scores.dt', self._std_perf_scores)

		# np.save(results_name_pre + 'best_params_index', best_params_index)
		# np.save(results_name_pre + 'best_params_pre.dt', best_params_out)
		# np.save(results_name_pre + 'best_params_in.dt', best_params_in)
		# np.save(results_name_pre + 'best_val_perf.dt', best_val_perf)
		# np.save(results_name_pre + 'best_val_std.dt', best_val_std)
		# np.save(results_name_pre + 'final_performance.dt', self._final_performance)
		# np.save(results_name_pre + 'final_confidence.dt', self._final_confidence)
		# np.save(results_name_pre + 'train_performance.dt', train_performance)
		# np.save(results_name_pre + 'train_std.dt', train_std)

		# np.save(results_name_pre + 'gram_matrix_time.dt', gram_matrix_time)
		# np.save(results_name_pre + 'average_gram_matrix_time.dt',
		#		 average_gram_matrix_time)
		# np.save(results_name_pre + 'std_gram_matrix_time.dt',
		#		 std_gram_matrix_time)
		# np.save(results_name_pre + 'best_gram_matrix_time.dt',
		#		 best_gram_matrix_time)


	def trial_do(self, param_list_pre_revised, param_list, gram_matrices, y, model_type, trial): # Test set level

	#	# get gram matrices from global variables.
	#	gram_matrices = np.reshape(G_gms.copy(), G_gms_shape, order='C')

		# Arrays to store scores
		train_pref = np.zeros((len(param_list_pre_revised), len(param_list)))
		val_pref = np.zeros((len(param_list_pre_revised), len(param_list)))
		test_pref = np.zeros((len(param_list_pre_revised), len(param_list)))

		# randomness added to seeds of split function below. "high" is "size" times
		# 10 so that at least 10 different random output will be yielded. Remove
		# these lines if identical outputs is required.
		rdm_out = np.random.RandomState(seed=None)
		rdm_seed_out_l = rdm_out.uniform(high=len(param_list_pre_revised) * 10,
									   size=len(param_list_pre_revised))
	#	print(trial, rdm_seed_out_l)
	#	print()
		# loop for each outer param tuple
		for index_out, params_out in enumerate(param_list_pre_revised):
			# get gram matrices from global variables.
	#		gm_now = G_gms[index_out * G_gms_shape[1] * G_gms_shape[2]:(index_out + 1) * G_gms_shape[1] * G_gms_shape[2]]
	#		gm_now = np.reshape(gm_now.copy(), (G_gms_shape[1], G_gms_shape[2]), order='C')
			gm_now = gram_matrices[index_out].copy()

			# split gram matrix and y to app and test sets.
			indices = range(len(y))
			# The argument "random_state" in function "train_test_split" can not be
			# set to None, because it will use RandomState instance used by
			# np.random, which is possible for multiple subprocesses to inherit the
			# same seed if they forked at the same time, leading to identical
			# random variates for different subprocesses. Instead, we use "trial"
			# and "index_out" parameters to generate different seeds for different
			# trials/subprocesses and outer loops. "rdm_seed_out_l" is used to add
			# randomness into seeds, so that it yields a different output every
			# time the program is run. To yield identical outputs every time,
			# remove the second line below. Same method is used to the "KFold"
			# function in the inner loop.
			rdm_seed_out = (trial + 1) * (index_out + 1)
			rdm_seed_out = (rdm_seed_out + int(rdm_seed_out_l[index_out])) % (2 ** 32 - 1)
	#		print(trial, rdm_seed_out)
			X_app, X_test, y_app, y_test, idx_app, idx_test = train_test_split(
				gm_now, y, indices, test_size=0.1,
				random_state=rdm_seed_out, shuffle=True)
	#		print(trial, idx_app, idx_test)
	#		print()
			X_app = X_app[:, idx_app]
			X_test = X_test[:, idx_app]
			y_app = np.array(y_app)
			y_test = np.array(y_test)

			rdm_seed_in_l = rdm_out.uniform(high=len(param_list) * 10,
									   size=len(param_list))
			# loop for each inner param tuple
			for index_in, params_in in enumerate(param_list):
	#			if trial == 0:
	#				print(index_out, index_in)
	#				print('params_in: ', params_in)
	#			st = time.time()
				rdm_seed_in = (trial + 1) * (index_out + 1) * (index_in + 1)
	#			print("rdm_seed_in1: ", trial, index_in, rdm_seed_in)
				rdm_seed_in = (rdm_seed_in + int(rdm_seed_in_l[index_in])) % (2 ** 32 - 1)
	#			print("rdm_seed_in2: ", trial, index_in, rdm_seed_in)
				inner_cv = KFold(n_splits=10, shuffle=True, random_state=rdm_seed_in)
				current_train_perf = []
				current_valid_perf = []
				current_test_perf = []

				# For regression use the Kernel Ridge method
	#			try:
				if self._model_type == 'regression':
					kr = KernelRidge(kernel='precomputed', **params_in)
					# loop for each split on validation set level
					# validation set level
					for train_index, valid_index in inner_cv.split(X_app):
	#					print("train_index, valid_index: ", trial, index_in, train_index, valid_index)
	#					if trial == 0:
	#						print('train_index: ', train_index)
	#						print('valid_index: ', valid_index)
	#						print('idx_test: ', idx_test)
	#						print('y_app[train_index]: ', y_app[train_index])
	#						print('X_app[train_index, :][:, train_index]: ', X_app[train_index, :][:, train_index])
	#						print('X_app[valid_index, :][:, train_index]: ', X_app[valid_index, :][:, train_index])
						kr.fit(X_app[train_index, :][:, train_index],
							   y_app[train_index])

						# predict on the train, validation and test set
						y_pred_train = kr.predict(
							X_app[train_index, :][:, train_index])
						y_pred_valid = kr.predict(
							X_app[valid_index, :][:, train_index])
	#					if trial == 0:
	#						print('y_pred_valid: ', y_pred_valid)
	#						print()
						y_pred_test = kr.predict(
							X_test[:, train_index])

						# root mean squared errors
						current_train_perf.append(
							np.sqrt(
								mean_squared_error(
									y_app[train_index], y_pred_train)))
						current_valid_perf.append(
							np.sqrt(
								mean_squared_error(
									y_app[valid_index], y_pred_valid)))
	#					if trial == 0:
	#						print(mean_squared_error(
	#								y_app[valid_index], y_pred_valid))
						current_test_perf.append(
							np.sqrt(
								mean_squared_error(
									y_test, y_pred_test)))
				# For clcassification use SVM
				else:
					svc = SVC(kernel='precomputed', cache_size=200,
							  verbose=False, **params_in)
					# loop for each split on validation set level
					# validation set level
					for train_index, valid_index in inner_cv.split(X_app):
	#						np.savez("bug.npy",X_app[train_index, :][:, train_index],y_app[train_index])
	#					if trial == 0:
	#						print('train_index: ', train_index)
	#						print('valid_index: ', valid_index)
	#						print('idx_test: ', idx_test)
	#						print('y_app[train_index]: ', y_app[train_index])
	#						print('X_app[train_index, :][:, train_index]: ', X_app[train_index, :][:, train_index])
	#						print('X_app[valid_index, :][:, train_index]: ', X_app[valid_index, :][:, train_index])
						svc.fit(X_app[train_index, :][:, train_index],
							   y_app[train_index])

						# predict on the train, validation and test set
						y_pred_train = svc.predict(
							X_app[train_index, :][:, train_index])
						y_pred_valid = svc.predict(
							X_app[valid_index, :][:, train_index])
						y_pred_test = svc.predict(
							X_test[:, train_index])

						# root mean squared errors
						current_train_perf.append(
							accuracy_score(y_app[train_index],
										   y_pred_train))
						current_valid_perf.append(
							accuracy_score(y_app[valid_index],
										   y_pred_valid))
						current_test_perf.append(
							accuracy_score(y_test, y_pred_test))
	#			except ValueError:
	#				print(sys.exc_info()[0])
	#				print(params_out, params_in)

				# average performance on inner splits
				train_pref[index_out][index_in] = np.mean(
					current_train_perf)
				val_pref[index_out][index_in] = np.mean(
					current_valid_perf)
				test_pref[index_out][index_in] = np.mean(
					current_test_perf)
	#			print(time.time() - st)
	#	if trial == 0:
	#		print('val_pref: ', val_pref)
	#		print('test_pref: ', test_pref)

		return train_pref, val_pref, test_pref


	def _parallel_trial_do(self, param_list_pre_revised, param_list, y, model_type, trial):
		train_pref, val_pref, test_pref = self._trial_do(param_list_pre_revised,
												   param_list, G_gms, y,
												   model_type, trial)
		return train_pref, val_pref, test_pref


	def printResultsInTable(self, param_list, param_list_pre_revised, average_val_scores,
							std_val_scores, average_perf_scores, std_perf_scores,
							average_train_scores, std_train_scores, gram_matrix_time,
							model_type, verbose):
		from collections import OrderedDict
		from tabulate import tabulate
		table_dict = {}
		if model_type == 'regression':
			for param_in in param_list:
				param_in['alpha'] = '{:.2e}'.format(param_in['alpha'])
		else:
			for param_in in param_list:
				param_in['C'] = '{:.2e}'.format(param_in['C'])
		table_dict['params'] = [{**param_out, **param_in}
								for param_in in param_list for param_out in param_list_pre_revised]
		table_dict['gram_matrix_time'] = [
			'{:.2f}'.format(gram_matrix_time[index_out])
			for param_in in param_list
			for index_out, _ in enumerate(param_list_pre_revised)
		]
		table_dict['valid_perf'] = [
			'{:.2f}±{:.2f}'.format(average_val_scores[index_out][index_in],
								   std_val_scores[index_out][index_in])
			for index_in, _ in enumerate(param_list)
			for index_out, _ in enumerate(param_list_pre_revised)
		]
		table_dict['test_perf'] = [
			'{:.2f}±{:.2f}'.format(average_perf_scores[index_out][index_in],
								   std_perf_scores[index_out][index_in])
			for index_in, _ in enumerate(param_list)
			for index_out, _ in enumerate(param_list_pre_revised)
		]
		table_dict['train_perf'] = [
			'{:.2f}±{:.2f}'.format(average_train_scores[index_out][index_in],
								   std_train_scores[index_out][index_in])
			for index_in, _ in enumerate(param_list)
			for index_out, _ in enumerate(param_list_pre_revised)
		]

		keyorder = [
			'params', 'train_perf', 'valid_perf', 'test_perf',
			'gram_matrix_time'
		]
		if verbose:
			print()
		tb_print = tabulate(OrderedDict(sorted(table_dict.items(),
							key=lambda i: keyorder.index(i[0]))), headers='keys')
	#			print(tb_print)
		return 'table of performance v.s. hyper-params:\n\n%s\n\n' % tb_print


"""
@author: linlin

@references: 

	[1] Thomas Gärtner, Peter Flach, and Stefan Wrobel. On graph kernels: 
	Hardness results and efficient alternatives. Learning Theory and Kernel
	Machines, pages 129–143, 2003.
"""

import sys
import time
from collections import Counter
from functools import partial

import networkx as nx
import numpy as np

from gklearn.utils.utils import direct_product
from gklearn.utils.graphdataset import get_dataset_attributes
from gklearn.utils.parallel import parallel_gm


def commonwalkkernel(*args,
					 node_label='atom',
					 edge_label='bond_type',
#					 n=None,
					 weight=1,
					 compute_method=None,
					 n_jobs=None,
					 chunksize=None,
					 verbose=True):
	"""Compute common walk graph kernels between graphs.

	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.
	
	G1, G2 : NetworkX graphs
		Two graphs between which the kernel is computed.
	node_label : string
		Node attribute used as symbolic label. The default node label is 'atom'.
	edge_label : string
		Edge attribute used as symbolic label. The default edge label is 'bond_type'.
	weight: integer
		Weight coefficient of different lengths of walks, which represents beta
		in 'exp' method and gamma in 'geo'.
	compute_method : string
		Method used to compute walk kernel. The Following choices are 
		available:

		'exp': method based on exponential serials applied on the direct 
		product graph, as shown in reference [1]. The time complexity is O(n^6) 
		for graphs with n vertices.

		'geo': method based on geometric serials applied on the direct product 
		graph, as shown in reference [1]. The time complexity is O(n^6) for 
		graphs with n vertices.

	n_jobs : int
		Number of jobs for parallelization. 

	Return
	------
	Kmatrix : Numpy matrix
		Kernel matrix, each element of which is a common walk kernel between 2 
		graphs.
	"""
#	n : integer
#		Longest length of walks. Only useful when applying the 'brute' method.
#		'brute': brute force, simply search for all walks and compare them.
	compute_method = compute_method.lower()
	# arrange all graphs in a list
	Gn = args[0] if len(args) == 1 else [args[0], args[1]]
	
	# remove graphs with only 1 node, as they do not have adjacency matrices 
	len_gn = len(Gn)
	Gn = [(idx, G) for idx, G in enumerate(Gn) if nx.number_of_nodes(G) != 1]
	idx = [G[0] for G in Gn]
	Gn = [G[1] for G in Gn]
	if len(Gn) != len_gn:
		if verbose:
			print('\n %d graphs are removed as they have only 1 node.\n' %
				  (len_gn - len(Gn)))
		
	ds_attrs = get_dataset_attributes(
		Gn,
		attr_names=['node_labeled', 'edge_labeled', 'is_directed'],
		node_label=node_label, edge_label=edge_label)
	if not ds_attrs['node_labeled']:
		for G in Gn:
			nx.set_node_attributes(G, '0', 'atom')
	if not ds_attrs['edge_labeled']:
		for G in Gn:
			nx.set_edge_attributes(G, '0', 'bond_type')
	if not ds_attrs['is_directed']:  #  convert
		Gn = [G.to_directed() for G in Gn]

	start_time = time.time()
	
	Kmatrix = np.zeros((len(Gn), len(Gn)))

	# ---- use pool.imap_unordered to parallel and track progress. ----
	def init_worker(gn_toshare):
		global G_gn
		G_gn = gn_toshare
	# direct product graph method - exponential
	if compute_method == 'exp':
		do_partial = partial(wrapper_cw_exp, node_label, edge_label, weight)
	# direct product graph method - geometric
	elif compute_method == 'geo':
		do_partial = partial(wrapper_cw_geo, node_label, edge_label, weight)  
	parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
				glbv=(Gn,), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose)  
	
	
#	pool = Pool(n_jobs)
#	itr = zip(combinations_with_replacement(Gn, 2),
#			  combinations_with_replacement(range(0, len(Gn)), 2))
#	len_itr = int(len(Gn) * (len(Gn) + 1) / 2)
#	if len_itr < 1000 * n_jobs:
#		chunksize = int(len_itr / n_jobs) + 1
#	else:
#		chunksize = 1000
#
#	# direct product graph method - exponential
#	if compute_method == 'exp':
#		do_partial = partial(wrapper_cw_exp, node_label, edge_label, weight)
#	# direct product graph method - geometric
#	elif compute_method == 'geo':
#		do_partial = partial(wrapper_cw_geo, node_label, edge_label, weight)
#
#	for i, j, kernel in tqdm(
#			pool.imap_unordered(do_partial, itr, chunksize),
#			desc='computing kernels',
#			file=sys.stdout):
#		Kmatrix[i][j] = kernel
#		Kmatrix[j][i] = kernel
#	pool.close()
#	pool.join()


#	# ---- direct running, normally use single CPU core. ----
#	# direct product graph method - exponential
#	itr = combinations_with_replacement(range(0, len(Gn)), 2)
#	if compute_method == 'exp':
#		for i, j in tqdm(itr, desc='Computing kernels', file=sys.stdout):
#			Kmatrix[i][j] = _commonwalkkernel_exp(Gn[i], Gn[j], node_label,
#													  edge_label, weight)
#			Kmatrix[j][i] = Kmatrix[i][j]
#
#	# direct product graph method - geometric
#	elif compute_method == 'geo':
#		for i, j in tqdm(itr, desc='Computing kernels', file=sys.stdout):
#			Kmatrix[i][j] = _commonwalkkernel_geo(Gn[i], Gn[j], node_label,
#													  edge_label, weight)
#			Kmatrix[j][i] = Kmatrix[i][j]


#	# search all paths use brute force.
#	elif compute_method == 'brute':
#		n = int(n)
#		# get all paths of all graphs before computing kernels to save time, but this may cost a lot of memory for large dataset.
#		all_walks = [
#			find_all_walks_until_length(Gn[i], n, node_label, edge_label)
#				for i in range(0, len(Gn))
#		]
#
#		for i in range(0, len(Gn)):
#			for j in range(i, len(Gn)):
#				Kmatrix[i][j] = _commonwalkkernel_brute(
#					all_walks[i],
#					all_walks[j],
#					node_label=node_label,
#					edge_label=edge_label)
#				Kmatrix[j][i] = Kmatrix[i][j]

	run_time = time.time() - start_time
	if verbose:
		print("\n --- kernel matrix of common walk kernel of size %d built in %s seconds ---"
			  % (len(Gn), run_time))

	return Kmatrix, run_time, idx


def _commonwalkkernel_exp(g1, g2, node_label, edge_label, beta):
	"""Compute walk graph kernels up to n between 2 graphs using exponential 
	series.

	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.
	node_label : string
		Node attribute used as label.
	edge_label : string
		Edge attribute used as label.
	beta : integer
		Weight.
	ij : tuple of integer
		Index of graphs between which the kernel is computed.

	Return
	------
	kernel : float
		The common walk Kernel between 2 graphs.
	"""

	# get tensor product / direct product
	gp = direct_product(g1, g2, node_label, edge_label)
	# return 0 if the direct product graph have no more than 1 node.
	if nx.number_of_nodes(gp) < 2:
		return 0
	A = nx.adjacency_matrix(gp).todense()
	# print(A)

	# from matplotlib import pyplot as plt
	# nx.draw_networkx(G1)
	# plt.show()
	# nx.draw_networkx(G2)
	# plt.show()
	# nx.draw_networkx(gp)
	# plt.show()
	# print(G1.nodes(data=True))
	# print(G2.nodes(data=True))
	# print(gp.nodes(data=True))
	# print(gp.edges(data=True))

	ew, ev = np.linalg.eig(A)
	# print('ew: ', ew)
	# print(ev)
	# T = np.matrix(ev)
	# print('T: ', T)
	# T = ev.I
	D = np.zeros((len(ew), len(ew)))
	for i in range(len(ew)):
		D[i][i] = np.exp(beta * ew[i])
		# print('D: ', D)
	# print('hshs: ', T.I * D * T)

	# print(np.exp(-2))
	# print(D)
	# print(np.exp(weight * D))
	# print(ev)
	# print(np.linalg.inv(ev))
	exp_D = ev * D * ev.T
	# print(exp_D)
	# print(np.exp(weight * A))
	# print('-------')

	return exp_D.sum()


def wrapper_cw_exp(node_label, edge_label, beta, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _commonwalkkernel_exp(G_gn[i], G_gn[j], node_label, edge_label, beta)


def _commonwalkkernel_geo(g1, g2, node_label, edge_label, gamma):
	"""Compute common walk graph kernels up to n between 2 graphs using 
	geometric series.

	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.
	node_label : string
		Node attribute used as label.
	edge_label : string
		Edge attribute used as label.
	gamma: integer
		Weight.
	ij : tuple of integer
		Index of graphs between which the kernel is computed.

	Return
	------
	kernel : float
		The common walk Kernel between 2 graphs.
	"""
	# get tensor product / direct product
	gp = direct_product(g1, g2, node_label, edge_label)
	# return 0 if the direct product graph have no more than 1 node.
	if nx.number_of_nodes(gp) < 2:
		return 0
	A = nx.adjacency_matrix(gp).todense()
	mat = np.identity(len(A)) - gamma * A
#	try:
	return mat.I.sum()
#	except np.linalg.LinAlgError:
#		return np.nan
	
	
def wrapper_cw_geo(node_label, edge_label, gama, itr):
	i = itr[0]
	j = itr[1]
	return i, j, _commonwalkkernel_geo(G_gn[i], G_gn[j], node_label, edge_label, gama)


def _commonwalkkernel_brute(walks1,
							walks2,
							node_label='atom',
							edge_label='bond_type',
							labeled=True):
	"""Compute walk graph kernels up to n between 2 graphs.

	Parameters
	----------
	walks1, walks2 : list
		List of walks in 2 graphs, where for unlabeled graphs, each walk is 
		represented by a list of nodes; while for labeled graphs, each walk is 
		represented by a string consists of labels of nodes and edges on that 
		walk.
	node_label : string
		node attribute used as label. The default node label is atom.
	edge_label : string
		edge attribute used as label. The default edge label is bond_type.
	labeled : boolean
		Whether the graphs are labeled. The default is True.

	Return
	------
	kernel : float
		Treelet Kernel between 2 graphs.
	"""
	counts_walks1 = dict(Counter(walks1))
	counts_walks2 = dict(Counter(walks2))
	all_walks = list(set(walks1 + walks2))

	vector1 = [(counts_walks1[walk] if walk in walks1 else 0)
			   for walk in all_walks]
	vector2 = [(counts_walks2[walk] if walk in walks2 else 0)
			   for walk in all_walks]
	kernel = np.dot(vector1, vector2)

	return kernel


# this method find walks repetively, it could be faster.
def find_all_walks_until_length(G,
								length,
								node_label='atom',
								edge_label='bond_type',
								labeled=True):
	"""Find all walks with a certain maximum length in a graph. 
	A recursive depth first search is applied.

	Parameters
	----------
	G : NetworkX graphs
		The graph in which walks are searched.
	length : integer
		The maximum length of walks.
	node_label : string
		node attribute used as label. The default node label is atom.
	edge_label : string
		edge attribute used as label. The default edge label is bond_type.
	labeled : boolean
		Whether the graphs are labeled. The default is True.

	Return
	------
	walk : list
		List of walks retrieved, where for unlabeled graphs, each walk is 
		represented by a list of nodes; while for labeled graphs, each walk 
		is represented by a string consists of labels of nodes and edges on 
		that walk.
	"""
	all_walks = []
	# @todo: in this way, the time complexity is close to N(d^n+d^(n+1)+...+1), which could be optimized to O(Nd^n)
	for i in range(0, length + 1):
		new_walks = find_all_walks(G, i)
		if new_walks == []:
			break
		all_walks.extend(new_walks)

	if labeled == True:  # convert paths to strings
		walk_strs = []
		for walk in all_walks:
			strlist = [
				G.node[node][node_label] +
				G[node][walk[walk.index(node) + 1]][edge_label]
				for node in walk[:-1]
			]
			walk_strs.append(''.join(strlist) + G.node[walk[-1]][node_label])

		return walk_strs

	return all_walks


def find_walks(G, source_node, length):
	"""Find all walks with a certain length those start from a source node. A 
	recursive depth first search is applied.

	Parameters
	----------
	G : NetworkX graphs
		The graph in which walks are searched.
	source_node : integer
		The number of the node from where all walks start.
	length : integer
		The length of walks.

	Return
	------
	walk : list of list
		List of walks retrieved, where each walk is represented by a list of 
		nodes.
	"""
	return [[source_node]] if length == 0 else \
		[[source_node] + walk for neighbor in G[source_node]
		 for walk in find_walks(G, neighbor, length - 1)]


def find_all_walks(G, length):
	"""Find all walks with a certain length in a graph. A recursive depth first
	search is applied.

	Parameters
	----------
	G : NetworkX graphs
		The graph in which walks are searched.
	length : integer
		The length of walks.

	Return
	------
	walk : list of list
		List of walks retrieved, where each walk is represented by a list of 
		nodes.
	"""
	all_walks = []
	for node in G:
		all_walks.extend(find_walks(G, node, length))

	# The following process is not carried out according to the original article
	# all_paths_r = [ path[::-1] for path in all_paths ]

	# # For each path, two presentation are retrieved from its two extremities. Remove one of them.
	# for idx, path in enumerate(all_paths[:-1]):
	#	 for path2 in all_paths_r[idx+1::]:
	#		 if path == path2:
	#			 all_paths[idx] = []
	#			 break

	# return list(filter(lambda a: a != [], all_paths))
	return all_walks



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 26 18:48:27 2020

@author: ljia
"""
import numpy as np
import networkx as nx
import os
from gklearn.dataset import DATASET_META, DataFetcher, DataLoader


class Dataset(object):


	def __init__(self, inputs=None, root='datasets', filename_targets=None, targets=None, mode='networkx', remove_null_graphs=True, clean_labels=True, reload=False, verbose=False, **kwargs):
		self._substructures = None
		self._node_label_dim = None
		self._edge_label_dim = None
		self._directed = None
		self._dataset_size = None
		self._total_node_num = None
		self._ave_node_num = None
		self._min_node_num = None
		self._max_node_num = None
		self._total_edge_num = None
		self._ave_edge_num = None
		self._min_edge_num = None
		self._max_edge_num = None
		self._ave_node_degree = None
		self._min_node_degree = None
		self._max_node_degree = None
		self._ave_fill_factor = None
		self._min_fill_factor = None
		self._max_fill_factor = None
		self._node_label_nums = None
		self._edge_label_nums = None
		self._node_attr_dim = None
		self._edge_attr_dim = None
		self._class_number = None
		self._ds_name = None
		self._task_type = None

		if inputs is None:
			self._graphs = None
			self._targets = None
			self._node_labels = None
			self._edge_labels = None
			self._node_attrs = None
			self._edge_attrs = None

		# If inputs is a list of graphs.
		elif isinstance(inputs, list):
			node_labels = kwargs.get('node_labels', None)
			node_attrs = kwargs.get('node_attrs', None)
			edge_labels = kwargs.get('edge_labels', None)
			edge_attrs = kwargs.get('edge_attrs', None)
			self.load_graphs(inputs, targets=targets)
			self.set_labels(node_labels=node_labels, node_attrs=node_attrs, edge_labels=edge_labels, edge_attrs=edge_attrs)
			if clean_labels:
				self.clean_labels()

		elif isinstance(inputs, str):
			# If inputs is predefined dataset name.
			if inputs in DATASET_META:
				self.load_predefined_dataset(inputs, root=root, clean_labels=clean_labels, reload=reload, verbose=verbose)
				self._ds_name = inputs

			# If the dataset is specially defined, i.g., Alkane_unlabeled, MAO_lite.
			elif self.is_special_dataset(inputs):
				self.load_special_dataset(inputs, root, clean_labels, reload, verbose)
				self._ds_name = inputs

			# If inputs is a file name.
			elif os.path.isfile(inputs):
				self.load_dataset(inputs, filename_targets=filename_targets, clean_labels=clean_labels, **kwargs)

			# If inputs is a file name.
			else:
				raise ValueError('The "inputs" argument "' + inputs + '" is not a valid dataset name or file name.')

		else:
			raise TypeError('The "inputs" argument cannot be recognized. "Inputs" can be a list of graphs, a predefined dataset name, or a file name of a dataset.')

		if remove_null_graphs:
			self.trim_dataset(edge_required=False)


	def load_dataset(self, filename, filename_targets=None, clean_labels=True, **kwargs):
		self._graphs, self._targets, label_names = DataLoader(filename, filename_targets=filename_targets, **kwargs).data
		self._node_labels = label_names['node_labels']
		self._node_attrs = label_names['node_attrs']
		self._edge_labels = label_names['edge_labels']
		self._edge_attrs = label_names['edge_attrs']
		if clean_labels:
			self.clean_labels()


	def load_graphs(self, graphs, targets=None):
		# this has to be followed by set_labels().
		self._graphs = graphs
		self._targets = targets
#		self.set_labels_attrs() # @todo


	def load_predefined_dataset(self, ds_name, root='datasets', clean_labels=True, reload=False, verbose=False):
		path = DataFetcher(name=ds_name, root=root, reload=reload, verbose=verbose).path

		if DATASET_META[ds_name]['database'] == 'tudataset':
			ds_file = os.path.join(path, ds_name + '_A.txt')
			fn_targets = None
		else:
			load_files = DATASET_META[ds_name]['load_files']
			if isinstance(load_files[0], str):
				ds_file = os.path.join(path, load_files[0])
			else: # load_files[0] is a list of files.
				ds_file = [os.path.join(path, fn) for fn in load_files[0]]
			fn_targets = os.path.join(path, load_files[1]) if len(load_files) == 2 else None

		# Get extra_params.
		if 'extra_params' in DATASET_META[ds_name]:
			kwargs = DATASET_META[ds_name]['extra_params']
		else:
			kwargs = {}

		# Get the task type that is associated with the dataset. If it is classification, get the number of classes.
		self._get_task_type(ds_name)


		self._graphs, self._targets, label_names = DataLoader(ds_file, filename_targets=fn_targets, **kwargs).data

		self._node_labels = label_names['node_labels']
		self._node_attrs = label_names['node_attrs']
		self._edge_labels = label_names['edge_labels']
		self._edge_attrs = label_names['edge_attrs']
		if clean_labels:
			self.clean_labels()

		# Deal with specific datasets.
		if ds_name == 'Alkane':
			self.trim_dataset(edge_required=True)
			self.remove_labels(node_labels=['atom_symbol'])


	def set_labels(self, node_labels=[], node_attrs=[], edge_labels=[], edge_attrs=[]):
		self._node_labels = node_labels
		self._node_attrs = node_attrs
		self._edge_labels = edge_labels
		self._edge_attrs = edge_attrs


	def set_labels_attrs(self, node_labels=None, node_attrs=None, edge_labels=None, edge_attrs=None):
		# @todo: remove labels which have only one possible values.
		if node_labels is None:
			self._node_labels = self._graphs[0].graph['node_labels']
#			# graphs are considered node unlabeled if all nodes have the same label.
#			infos.update({'node_labeled': is_nl if node_label_num > 1 else False})
		if node_attrs is None:
			self._node_attrs = self._graphs[0].graph['node_attrs']
#		for G in Gn:
#			for n in G.nodes(data=True):
#				if 'attributes' in n[1]:
#					return len(n[1]['attributes'])
#		return 0
		if edge_labels is None:
			self._edge_labels = self._graphs[0].graph['edge_labels']
#			# graphs are considered edge unlabeled if all edges have the same label.
#			infos.update({'edge_labeled': is_el if edge_label_num > 1 else False})
		if edge_attrs is None:
			self._edge_attrs = self._graphs[0].graph['edge_attrs']
#		for G in Gn:
#			if nx.number_of_edges(G) > 0:
#				for e in G.edges(data=True):
#					if 'attributes' in e[2]:
#						return len(e[2]['attributes'])
#		return 0


	def get_dataset_infos(self, keys=None, params=None):
		"""Computes and returns the structure and property information of the graph dataset.

		Parameters
		----------
		keys : list, optional
			A list of strings which indicate which informations will be returned. The
			possible choices includes:

			'substructures': sub-structures graphs contains, including 'linear', 'non
		linear' and 'cyclic'.

			'node_label_dim': whether vertices have symbolic labels.

			'edge_label_dim': whether egdes have symbolic labels.

			'directed': whether graphs in dataset are directed.

			'dataset_size': number of graphs in dataset.

			'total_node_num': total number of vertices of all graphs in dataset.

			'ave_node_num': average number of vertices of graphs in dataset.

			'min_node_num': minimum number of vertices of graphs in dataset.

			'max_node_num': maximum number of vertices of graphs in dataset.

			'total_edge_num': total number of edges of all graphs in dataset.

			'ave_edge_num': average number of edges of graphs in dataset.

			'min_edge_num': minimum number of edges of graphs in dataset.

			'max_edge_num': maximum number of edges of graphs in dataset.

			'ave_node_degree': average vertex degree of graphs in dataset.

			'min_node_degree': minimum vertex degree of graphs in dataset.

			'max_node_degree': maximum vertex degree of graphs in dataset.

			'ave_fill_factor': average fill factor (number_of_edges /
		(number_of_nodes ** 2)) of graphs in dataset.

			'min_fill_factor': minimum fill factor of graphs in dataset.

			'max_fill_factor': maximum fill factor of graphs in dataset.

			'node_label_nums': list of numbers of symbolic vertex labels of graphs in dataset.

			'edge_label_nums': list number of symbolic edge labels of graphs in dataset.

			'node_attr_dim': number of dimensions of non-symbolic vertex labels.
		Extracted from the 'attributes' attribute of graph nodes.

			'edge_attr_dim': number of dimensions of non-symbolic edge labels.
		Extracted from the 'attributes' attribute of graph edges.

			'class_number': number of classes. Only available for classification problems.

			'all_degree_entropy': the entropy of degree distribution of each graph.

			'ave_degree_entropy': the average entropy of degree distribution of all graphs.

			All informations above will be returned if `keys` is not given.

		params: dict of dict, optional
			A dictinary which contains extra parameters for each possible
			element in ``keys``.

		Return
		------
		dict
			Information of the graph dataset keyed by `keys`.
		"""
		infos = {}

		if keys == None:
			keys = [
				'substructures',
				'node_label_dim',
				'edge_label_dim',
				'directed',
				'dataset_size',
				'total_node_num',
				'ave_node_num',
				'min_node_num',
				'max_node_num',
				'total_edge_num',
				'ave_edge_num',
				'min_edge_num',
				'max_edge_num',
				'ave_node_degree',
				'min_node_degree',
				'max_node_degree',
				'ave_fill_factor',
				'min_fill_factor',
				'max_fill_factor',
				'node_label_nums',
				'edge_label_nums',
				'node_attr_dim',
				'edge_attr_dim',
				'class_number',
				'all_degree_entropy',
				'ave_degree_entropy',
				'class_type'
			]

		# dataset size
		if 'dataset_size' in keys:
			if self._dataset_size is None:
				self._dataset_size = self._get_dataset_size()
			infos['dataset_size'] = self._dataset_size

		# graph node number
		if any(i in keys for i in ['total_node_num', 'ave_node_num', 'min_node_num', 'max_node_num']):
			all_node_nums = self._get_all_node_nums()

		if 'total_node_num' in keys:
			if self._total_node_num is None:
				self._total_node_num = self._get_total_node_num(all_node_nums)
			infos['total_node_num'] = self._total_node_num

		if 'ave_node_num' in keys:
			if self._ave_node_num is None:
				self._ave_node_num = self._get_ave_node_num(all_node_nums)
			infos['ave_node_num'] = self._ave_node_num

		if 'min_node_num' in keys:
			if self._min_node_num is None:
				self._min_node_num = self._get_min_node_num(all_node_nums)
			infos['min_node_num'] = self._min_node_num

		if 'max_node_num' in keys:
			if self._max_node_num is None:
				self._max_node_num = self._get_max_node_num(all_node_nums)
			infos['max_node_num'] = self._max_node_num

		# graph edge number
		if any(i in keys for i in ['total_edge_num', 'ave_edge_num', 'min_edge_num', 'max_edge_num']):
			all_edge_nums = self._get_all_edge_nums()

		if 'total_edge_num' in keys:
			if self._total_edge_num is None:
				self._total_edge_num = self._get_total_edge_num(all_edge_nums)
			infos['total_edge_num'] = self._total_edge_num

		if 'ave_edge_num' in keys:
			if self._ave_edge_num is None:
				self._ave_edge_num = self._get_ave_edge_num(all_edge_nums)
			infos['ave_edge_num'] = self._ave_edge_num

		if 'max_edge_num' in keys:
			if self._max_edge_num is None:
				self._max_edge_num = self._get_max_edge_num(all_edge_nums)
			infos['max_edge_num'] = self._max_edge_num

		if 'min_edge_num' in keys:
			if self._min_edge_num is None:
				self._min_edge_num = self._get_min_edge_num(all_edge_nums)
			infos['min_edge_num'] = self._min_edge_num

		# label number
		if 'node_label_dim' in keys:
			if self._node_label_dim is None:
				self._node_label_dim = self._get_node_label_dim()
			infos['node_label_dim'] = self._node_label_dim

		if 'node_label_nums' in keys:
			if self._node_label_nums is None:
				self._node_label_nums = {}
				for node_label in self._node_labels:
					self._node_label_nums[node_label] = self._get_node_label_num(node_label)
			infos['node_label_nums'] = self._node_label_nums

		if 'edge_label_dim' in keys:
			if self._edge_label_dim is None:
				self._edge_label_dim = self._get_edge_label_dim()
			infos['edge_label_dim'] = self._edge_label_dim

		if 'edge_label_nums' in keys:
			if self._edge_label_nums is None:
				self._edge_label_nums = {}
				for edge_label in self._edge_labels:
					self._edge_label_nums[edge_label] = self._get_edge_label_num(edge_label)
			infos['edge_label_nums'] = self._edge_label_nums

		if 'directed' in keys or 'substructures' in keys:
			if self._directed is None:
				self._directed = self._is_directed()
			infos['directed'] = self._directed

		# node degree
		if any(i in keys for i in ['ave_node_degree', 'max_node_degree', 'min_node_degree']):
			all_node_degrees = self._get_all_node_degrees()

		if 'ave_node_degree' in keys:
			if self._ave_node_degree is None:
				self._ave_node_degree = self._get_ave_node_degree(all_node_degrees)
			infos['ave_node_degree'] = self._ave_node_degree

		if 'max_node_degree' in keys:
			if self._max_node_degree is None:
				self._max_node_degree = self._get_max_node_degree(all_node_degrees)
			infos['max_node_degree'] = self._max_node_degree

		if 'min_node_degree' in keys:
			if self._min_node_degree is None:
				self._min_node_degree = self._get_min_node_degree(all_node_degrees)
			infos['min_node_degree'] = self._min_node_degree

		# fill factor
		if any(i in keys for i in ['ave_fill_factor', 'max_fill_factor', 'min_fill_factor']):
			all_fill_factors = self._get_all_fill_factors()

		if 'ave_fill_factor' in keys:
			if self._ave_fill_factor is None:
				self._ave_fill_factor = self._get_ave_fill_factor(all_fill_factors)
			infos['ave_fill_factor'] = self._ave_fill_factor

		if 'max_fill_factor' in keys:
			if self._max_fill_factor is None:
				self._max_fill_factor = self._get_max_fill_factor(all_fill_factors)
			infos['max_fill_factor'] = self._max_fill_factor

		if 'min_fill_factor' in keys:
			if self._min_fill_factor is None:
				self._min_fill_factor = self._get_min_fill_factor(all_fill_factors)
			infos['min_fill_factor'] = self._min_fill_factor

		if 'substructures' in keys:
			if self._substructures is None:
				self._substructures = self._get_substructures()
			infos['substructures'] = self._substructures

		if 'class_number' in keys:
			if self._class_number is None:
				self._class_number = self._get_class_num()
			infos['class_number'] = self._class_number

		if 'node_attr_dim' in keys:
			if self._node_attr_dim is None:
				self._node_attr_dim = self._get_node_attr_dim()
			infos['node_attr_dim'] = self._node_attr_dim

		if 'edge_attr_dim' in keys:
			if self._edge_attr_dim is None:
				self._edge_attr_dim = self._get_edge_attr_dim()
			infos['edge_attr_dim'] = self._edge_attr_dim

		# entropy of degree distribution.

		if 'all_degree_entropy' in keys:
			if params is not None and ('all_degree_entropy' in params) and ('base' in params['all_degree_entropy']):
				base = params['all_degree_entropy']['base']
			else:
				base = None
			infos['all_degree_entropy'] = self._compute_all_degree_entropy(base=base)

		if 'ave_degree_entropy' in keys:
			if params is not None and ('ave_degree_entropy' in params) and ('base' in params['ave_degree_entropy']):
				base = params['ave_degree_entropy']['base']
			else:
				base = None
			infos['ave_degree_entropy'] = np.mean(self._compute_all_degree_entropy(base=base))

		if 'task_type' in keys:
			if self._task_type is None:
				self._task_type = self._get_task_type()
			infos['task_type'] = self._task_type

		return infos


	def print_graph_infos(self, infos):
		from collections import OrderedDict
		keys = list(infos.keys())
		print(OrderedDict(sorted(infos.items(), key=lambda i: keys.index(i[0]))))


	def remove_labels(self, node_labels=[], edge_labels=[], node_attrs=[], edge_attrs=[]):
		node_labels = [item for item in node_labels if item in self._node_labels]
		edge_labels = [item for item in edge_labels if item in self._edge_labels]
		node_attrs = [item for item in node_attrs if item in self._node_attrs]
		edge_attrs = [item for item in edge_attrs if item in self._edge_attrs]

		for g in self._graphs:
			for nd in g.nodes():
				for nl in node_labels:
					del g.nodes[nd][nl]
				for na in node_attrs:
					del g.nodes[nd][na]
			for ed in g.edges():
				for el in edge_labels:
					del g.edges[ed][el]
				for ea in edge_attrs:
					del g.edges[ed][ea]
		if len(node_labels) > 0:
			self._node_labels = [nl for nl in self._node_labels if nl not in node_labels]
		if len(edge_labels) > 0:
			self._edge_labels = [el for el in self._edge_labels if el not in edge_labels]
		if len(node_attrs) > 0:
			self._node_attrs = [na for na in self._node_attrs if na not in node_attrs]
		if len(edge_attrs) > 0:
			self._edge_attrs = [ea for ea in self._edge_attrs if ea not in edge_attrs]


	def clean_labels(self):
		labels = []
		for name in self._node_labels:
			label = set()
			for G in self._graphs:
				label = label | set(nx.get_node_attributes(G, name).values())
				if len(label) > 1:
					labels.append(name)
					break
			if len(label) < 2:
				for G in self._graphs:
					for nd in G.nodes():
						del G.nodes[nd][name]
		self._node_labels = labels

		labels = []
		for name in self._edge_labels:
			label = set()
			for G in self._graphs:
				label = label | set(nx.get_edge_attributes(G, name).values())
				if len(label) > 1:
					labels.append(name)
					break
			if len(label) < 2:
				for G in self._graphs:
					for ed in G.edges():
						del G.edges[ed][name]
		self._edge_labels = labels

		labels = []
		for name in self._node_attrs:
			label = set()
			for G in self._graphs:
				label = label | set(nx.get_node_attributes(G, name).values())
				if len(label) > 1:
					labels.append(name)
					break
			if len(label) < 2:
				for G in self._graphs:
					for nd in G.nodes():
						del G.nodes[nd][name]
		self._node_attrs = labels

		labels = []
		for name in self._edge_attrs:
			label = set()
			for G in self._graphs:
				label = label | set(nx.get_edge_attributes(G, name).values())
				if len(label) > 1:
					labels.append(name)
					break
			if len(label) < 2:
				for G in self._graphs:
					for ed in G.edges():
						del G.edges[ed][name]
		self._edge_attrs = labels


	def cut_graphs(self, range_):
		self._graphs = [self._graphs[i] for i in range_]
		if self._targets is not None:
			self._targets = [self._targets[i] for i in range_]
		self.clean_labels()


	def trim_dataset(self, edge_required=False):
		if edge_required: # @todo: there is a possibility that some node labels will be removed.
			trimed_pairs = [(idx, g) for idx, g in enumerate(self._graphs) if (nx.number_of_nodes(g) != 0 and nx.number_of_edges(g) != 0)]
		else:
			trimed_pairs = [(idx, g) for idx, g in enumerate(self._graphs) if nx.number_of_nodes(g) != 0]
		idx = [p[0] for p in trimed_pairs]
		self._graphs = [p[1] for p in trimed_pairs]
		self._targets = [self._targets[i] for i in idx]
		self.clean_labels()


	def copy(self):
		dataset = Dataset()
		graphs = [g.copy() for g in self._graphs] if self._graphs is not None else None
		target = self._targets.copy() if self._targets is not None else None
		node_labels = self._node_labels.copy() if self._node_labels is not None else None
		node_attrs = self._node_attrs.copy() if self._node_attrs is not None else None
		edge_labels = self._edge_labels.copy() if self._edge_labels is not None else None
		edge_attrs = self._edge_attrs.copy() if self._edge_attrs is not None else None
		dataset.load_graphs(graphs, target)
		dataset.set_labels(node_labels=node_labels, node_attrs=node_attrs, edge_labels=edge_labels, edge_attrs=edge_attrs)
		# @todo: clean_labels and add other class members?
		return dataset


	def is_special_dataset(self, inputs):
		if inputs.endswith('_unlabeled'):
			return True
		if inputs == 'MAO_lite':
			return True
		if inputs == 'Monoterpens':
			return True
		return False


	def load_special_dataset(self, inputs, root, clean_labels, reload, verbose):
		if inputs.endswith('_unlabeled'):
			self.load_predefined_dataset(inputs[:len(inputs) - 10], root=root, clean_labels=clean_labels, reload=reload, verbose=verbose)

			self.remove_labels(node_labels=self._node_labels,
					  edge_labels=self._edge_labels,
					  node_attrs=self._node_attrs,
					  edge_attrs=self._edge_attrs)

		elif inputs == 'MAO_lite':
			self.load_predefined_dataset(inputs[:len(inputs) - 5], root=root, clean_labels=clean_labels, reload=reload, verbose=verbose)

			self.remove_labels(edge_labels=['bond_stereo'], node_attrs=['x', 'y'])

		elif inputs == 'Monoterpens':
			self.load_predefined_dataset('Monoterpenoides', root=root, clean_labels=clean_labels, reload=reload, verbose=verbose)


	def get_all_node_labels(self):
		node_labels = []
		for g in self._graphs:
			for n in g.nodes():
				nl = tuple(g.nodes[n].items())
				if nl not in node_labels:
					node_labels.append(nl)
		return node_labels


	def get_all_edge_labels(self):
		edge_labels = []
		for g in self._graphs:
			for e in g.edges():
				el = tuple(g.edges[e].items())
				if el not in edge_labels:
					edge_labels.append(el)
		return edge_labels


	def _get_dataset_size(self):
		return len(self._graphs)


	def _get_all_node_nums(self):
		return [nx.number_of_nodes(G) for G in self._graphs]


	def _get_total_node_nums(self, all_node_nums):
		return np.sum(all_node_nums)


	def _get_ave_node_num(self, all_node_nums):
		return np.mean(all_node_nums)


	def _get_min_node_num(self, all_node_nums):
		return np.amin(all_node_nums)


	def _get_max_node_num(self, all_node_nums):
		return np.amax(all_node_nums)


	def _get_all_edge_nums(self):
		return [nx.number_of_edges(G) for G in self._graphs]


	def _get_total_edge_nums(self, all_edge_nums):
		return np.sum(all_edge_nums)


	def _get_ave_edge_num(self, all_edge_nums):
		return np.mean(all_edge_nums)


	def _get_min_edge_num(self, all_edge_nums):
		return np.amin(all_edge_nums)


	def _get_max_edge_num(self, all_edge_nums):
		return np.amax(all_edge_nums)


	def _get_node_label_dim(self):
		return len(self._node_labels)


	def _get_node_label_num(self, node_label):
		nl = set()
		for G in self._graphs:
			nl = nl | set(nx.get_node_attributes(G, node_label).values())
		return len(nl)


	def _get_edge_label_dim(self):
		return len(self._edge_labels)


	def _get_edge_label_num(self, edge_label):
		el = set()
		for G in self._graphs:
			el = el | set(nx.get_edge_attributes(G, edge_label).values())
		return len(el)


	def _is_directed(self):
		return nx.is_directed(self._graphs[0])


	def _get_all_node_degrees(self):
		return [np.mean(list(dict(G.degree()).values())) for G in self._graphs]


	def _get_ave_node_degree(self, all_node_degrees):
		return np.mean(all_node_degrees)


	def _get_max_node_degree(self, all_node_degrees):
		return np.amax(all_node_degrees)


	def _get_min_node_degree(self, all_node_degrees):
		return np.amin(all_node_degrees)


	def _get_all_fill_factors(self):
		"""Get fill factor, the number of non-zero entries in the adjacency matrix.

		Returns
		-------
		list[float]
			List of fill factors for all graphs.
		"""
		return [nx.number_of_edges(G) / (nx.number_of_nodes(G) ** 2) for G in self._graphs]


	def _get_ave_fill_factor(self, all_fill_factors):
		return np.mean(all_fill_factors)


	def _get_max_fill_factor(self, all_fill_factors):
		return np.amax(all_fill_factors)


	def _get_min_fill_factor(self, all_fill_factors):
		return np.amin(all_fill_factors)


	def _get_substructures(self):
		subs = set()
		for G in self._graphs:
			degrees = list(dict(G.degree()).values())
			if any(i == 2 for i in degrees):
				subs.add('linear')
			if np.amax(degrees) >= 3:
				subs.add('non linear')
			if 'linear' in subs and 'non linear' in subs:
				break

		if self._directed:
			for G in self._graphs:
				if len(list(nx.find_cycle(G))) > 0:
					subs.add('cyclic')
					break
			# else:
			#	 # @todo: this method does not work for big graph with large amount of edges like D&D, try a better way.
			#	 upper = np.amin([nx.number_of_edges(G) for G in Gn]) * 2 + 10
			#	 for G in Gn:
			#		 if (nx.number_of_edges(G) < upper):
			#			 cyc = list(nx.simple_cycles(G.to_directed()))
			#			 if any(len(i) > 2 for i in cyc):
			#				 subs.add('cyclic')
			#				 break
			#	 if 'cyclic' not in subs:
			#		 for G in Gn:
			#			 cyc = list(nx.simple_cycles(G.to_directed()))
			#			 if any(len(i) > 2 for i in cyc):
			#				 subs.add('cyclic')
			#				 break

			return subs


	def _get_class_num(self):
		return len(set(self._targets))


	def _get_node_attr_dim(self):
		return len(self._node_attrs)


	def _get_edge_attr_dim(self):
		return len(self._edge_attrs)


	def _compute_all_degree_entropy(self, base=None):
		"""Compute the entropy of degree distribution of each graph.

		Parameters
		----------
		base : float, optional
			The logarithmic base to use. The default is ``e`` (natural logarithm).

		Returns
		-------
		degree_entropy : float
			The calculated entropy.
		"""
		from gklearn.utils.stats import entropy

		degree_entropy = []
		for g in self._graphs:
			degrees = list(dict(g.degree()).values())
			en = entropy(degrees, base=base)
			degree_entropy.append(en)
		return degree_entropy


	def _get_task_type(self, ds_name):
		if 'task_type' in DATASET_META[ds_name]:
			self._task_type = DATASET_META[ds_name]['task_type']
		if self._task_type == 'classification' and self._class_number is None and 'class_number' in DATASET_META[ds_name]:
				self._class_number = DATASET_META[ds_name]['class_number']


	@property
	def graphs(self):
		return self._graphs


	@property
	def targets(self):
		return self._targets


	@property
	def node_labels(self):
		return self._node_labels


	@property
	def edge_labels(self):
		return self._edge_labels


	@property
	def node_attrs(self):
		return self._node_attrs


	@property
	def edge_attrs(self):
		return self._edge_attrs


def split_dataset_by_target(dataset):
	from gklearn.preimage.utils import get_same_item_indices

	graphs = dataset.graphs
	targets = dataset.targets
	datasets = []
	idx_targets = get_same_item_indices(targets)
	for key, val in idx_targets.items():
		sub_graphs = [graphs[i] for i in val]
		sub_dataset = Dataset()
		sub_dataset.load_graphs(sub_graphs, [key] * len(val))
		node_labels = dataset.node_labels.copy() if dataset.node_labels is not None else None
		node_attrs = dataset.node_attrs.copy() if dataset.node_attrs is not None else None
		edge_labels = dataset.edge_labels.copy() if dataset.edge_labels is not None else None
		edge_attrs = dataset.edge_attrs.copy() if dataset.edge_attrs is not None else None
		sub_dataset.set_labels(node_labels=node_labels, node_attrs=node_attrs, edge_labels=edge_labels, edge_attrs=edge_attrs)
		datasets.append(sub_dataset)
		# @todo: clean_labels?
	return datasets


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 17 19:05:07 2019

Useful functions.
@author: ljia
"""
#import networkx as nx

import multiprocessing
import numpy as np

from gklearn.kernels.marginalizedKernel import marginalizedkernel
from gklearn.kernels.untilHPathKernel import untilhpathkernel
from gklearn.kernels.spKernel import spkernel
import functools
from gklearn.utils.kernels import deltakernel, gaussiankernel, kernelproduct, polynomialkernel
from gklearn.kernels.structuralspKernel import structuralspkernel
from gklearn.kernels.treeletKernel import treeletkernel
from gklearn.kernels.weisfeilerLehmanKernel import weisfeilerlehmankernel
from gklearn.utils import Dataset
import csv
import networkx as nx
import os


def generate_median_preimages_by_class(ds_name, mpg_options, kernel_options, ged_options, mge_options, save_results=True, save_medians=True, plot_medians=True, load_gm='auto', dir_save='', irrelevant_labels=None, edge_required=False, cut_range=None):
	import os.path
	from gklearn.preimage import MedianPreimageGenerator
	from gklearn.utils import split_dataset_by_target
	from gklearn.utils.graphfiles import saveGXL
	
	# 1. get dataset.
	print('1. getting dataset...')
	dataset_all = Dataset()
	dataset_all.load_predefined_dataset(ds_name)
	dataset_all.trim_dataset(edge_required=edge_required)
	if irrelevant_labels is not None:
		dataset_all.remove_labels(**irrelevant_labels)
	if cut_range is not None:
		dataset_all.cut_graphs(cut_range)
	datasets = split_dataset_by_target(dataset_all)

	if save_results:
		# create result files.
		print('creating output files...')
		fn_output_detail, fn_output_summary = _init_output_file_preimage(ds_name, kernel_options['name'], mpg_options['fit_method'], dir_save)
		
	sod_sm_list = []
	sod_gm_list = []
	dis_k_sm_list = []
	dis_k_gm_list = []
	dis_k_gi_min_list = []
	time_optimize_ec_list = []
	time_generate_list = []
	time_total_list = []
	itrs_list = []
	converged_list = []
	num_updates_ecc_list = []
	mge_decrease_order_list = []
	mge_increase_order_list = []
	mge_converged_order_list = []
	nb_sod_sm2gm = [0, 0, 0]
	nb_dis_k_sm2gm = [0, 0, 0]
	nb_dis_k_gi2sm = [0, 0, 0]
	nb_dis_k_gi2gm = [0, 0, 0]
	dis_k_max_list = []	
	dis_k_min_list = []
	dis_k_mean_list = []
	if load_gm == 'auto':
		gm_fname = dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm.npz'
		gmfile_exist = os.path.isfile(os.path.abspath(gm_fname))
		if gmfile_exist:
			gmfile = np.load(gm_fname, allow_pickle=True) # @todo: may not be safe.
			gram_matrix_unnorm_list = [item for item in gmfile['gram_matrix_unnorm_list']]
			time_precompute_gm_list = gmfile['run_time_list'].tolist()
		else:
			gram_matrix_unnorm_list = []
			time_precompute_gm_list = []
	elif not load_gm:
		gram_matrix_unnorm_list = []
		time_precompute_gm_list = []
	else:
		gm_fname = dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm.npz'
		gmfile = np.load(gm_fname, allow_pickle=True) # @todo: may not be safe.
		gram_matrix_unnorm_list = [item for item in gmfile['gram_matrix_unnorm_list']]
		time_precompute_gm_list = gmfile['run_time_list'].tolist()
#	repeats_better_sod_sm2gm = []
#	repeats_better_dis_k_sm2gm = []
#	repeats_better_dis_k_gi2sm = []
#	repeats_better_dis_k_gi2gm = []		
		
	print('starting generating preimage for each class of target...')
	idx_offset = 0
	for idx, dataset in enumerate(datasets):
		target = dataset.targets[0]
		print('\ntarget =', target, '\n')
#		if target != 1:
# 			continue
		
		num_graphs = len(dataset.graphs)
		if num_graphs < 2:
			print('\nnumber of graphs = ', num_graphs, ', skip.\n')
			idx_offset += 1
			continue
			
		# 2. set parameters.
		print('2. initializing mpg and setting parameters...')
		if load_gm:
			if gmfile_exist:
				mpg_options['gram_matrix_unnorm'] = gram_matrix_unnorm_list[idx - idx_offset]
				mpg_options['runtime_precompute_gm'] = time_precompute_gm_list[idx - idx_offset]
		mpg = MedianPreimageGenerator()
		mpg.dataset = dataset
		mpg.set_options(**mpg_options.copy())
		mpg.kernel_options = kernel_options.copy()
		mpg.ged_options = ged_options.copy()
		mpg.mge_options = mge_options.copy()

		# 3. compute median preimage.
		print('3. computing median preimage...')
		mpg.run()
		results = mpg.get_results()
		
		# 4. compute pairwise kernel distances.
		print('4. computing pairwise kernel distances...')
		_, dis_k_max, dis_k_min, dis_k_mean = mpg.graph_kernel.compute_distance_matrix()
		dis_k_max_list.append(dis_k_max)
		dis_k_min_list.append(dis_k_min)
		dis_k_mean_list.append(dis_k_mean)
		
		# 5. save results (and median graphs).
		print('5. saving results (and median graphs)...')
		# write result detail.
		if save_results:
			print('writing results to files...')
			sod_sm2gm = get_relations(np.sign(results['sod_gen_median'] - results['sod_set_median']))
			dis_k_sm2gm = get_relations(np.sign(results['k_dis_gen_median'] - results['k_dis_set_median']))
			dis_k_gi2sm = get_relations(np.sign(results['k_dis_set_median'] - results['k_dis_dataset']))
			dis_k_gi2gm = get_relations(np.sign(results['k_dis_gen_median'] - results['k_dis_dataset']))

			f_detail = open(dir_save + fn_output_detail, 'a')
			csv.writer(f_detail).writerow([ds_name, kernel_options['name'], 
					  ged_options['edit_cost'], ged_options['method'], 
					  ged_options['attr_distance'], mpg_options['fit_method'], 
					  num_graphs, target, 1, 
					  results['sod_set_median'], results['sod_gen_median'],
					  results['k_dis_set_median'], results['k_dis_gen_median'], 
					  results['k_dis_dataset'], sod_sm2gm, dis_k_sm2gm, 
					  dis_k_gi2sm, dis_k_gi2gm,	results['edit_cost_constants'],
					  results['runtime_precompute_gm'], results['runtime_optimize_ec'], 
					  results['runtime_generate_preimage'], results['runtime_total'],
					  results['itrs'], results['converged'],
					  results['num_updates_ecc'],
					  results['mge']['num_decrease_order'] > 0, # @todo: not suitable for multi-start mge
					  results['mge']['num_increase_order'] > 0,
					  results['mge']['num_converged_descents'] > 0])
			f_detail.close()
		
			# compute result summary.
			sod_sm_list.append(results['sod_set_median'])
			sod_gm_list.append(results['sod_gen_median'])
			dis_k_sm_list.append(results['k_dis_set_median'])
			dis_k_gm_list.append(results['k_dis_gen_median'])
			dis_k_gi_min_list.append(results['k_dis_dataset'])
			time_precompute_gm_list.append(results['runtime_precompute_gm'])
			time_optimize_ec_list.append(results['runtime_optimize_ec'])
			time_generate_list.append(results['runtime_generate_preimage'])
			time_total_list.append(results['runtime_total'])
			itrs_list.append(results['itrs'])
			converged_list.append(results['converged'])
			num_updates_ecc_list.append(results['num_updates_ecc'])
			mge_decrease_order_list.append(results['mge']['num_decrease_order'] > 0)
			mge_increase_order_list.append(results['mge']['num_increase_order'] > 0)
			mge_converged_order_list.append(results['mge']['num_converged_descents'] > 0)
			# # SOD SM -> GM
			if results['sod_set_median'] > results['sod_gen_median']:
				nb_sod_sm2gm[0] += 1
	#			repeats_better_sod_sm2gm.append(1)
			elif results['sod_set_median'] == results['sod_gen_median']:
				nb_sod_sm2gm[1] += 1
			elif results['sod_set_median'] < results['sod_gen_median']:
				nb_sod_sm2gm[2] += 1
			# # dis_k SM -> GM
			if results['k_dis_set_median'] > results['k_dis_gen_median']:
				nb_dis_k_sm2gm[0] += 1
	#			repeats_better_dis_k_sm2gm.append(1)
			elif results['k_dis_set_median'] == results['k_dis_gen_median']:
				nb_dis_k_sm2gm[1] += 1
			elif results['k_dis_set_median'] < results['k_dis_gen_median']:
				nb_dis_k_sm2gm[2] += 1
			# # dis_k gi -> SM
			if results['k_dis_dataset'] > results['k_dis_set_median']:
				nb_dis_k_gi2sm[0] += 1
	#			repeats_better_dis_k_gi2sm.append(1)
			elif results['k_dis_dataset'] == results['k_dis_set_median']:
				nb_dis_k_gi2sm[1] += 1
			elif results['k_dis_dataset'] < results['k_dis_set_median']:
				nb_dis_k_gi2sm[2] += 1
			# # dis_k gi -> GM
			if results['k_dis_dataset'] > results['k_dis_gen_median']:
				nb_dis_k_gi2gm[0] += 1
	#			repeats_better_dis_k_gi2gm.append(1)
			elif results['k_dis_dataset'] == results['k_dis_gen_median']:
				nb_dis_k_gi2gm[1] += 1
			elif results['k_dis_dataset'] < results['k_dis_gen_median']:
				nb_dis_k_gi2gm[2] += 1

			# write result summary for each letter.
			f_summary = open(dir_save + fn_output_summary, 'a')
			csv.writer(f_summary).writerow([ds_name, kernel_options['name'], 
					  ged_options['edit_cost'], ged_options['method'], 
					  ged_options['attr_distance'], mpg_options['fit_method'], 
					  num_graphs, target,
					  results['sod_set_median'], results['sod_gen_median'],
					  results['k_dis_set_median'], results['k_dis_gen_median'], 
					  results['k_dis_dataset'], sod_sm2gm, dis_k_sm2gm, 
					  dis_k_gi2sm, dis_k_gi2gm, 
					  results['runtime_precompute_gm'], results['runtime_optimize_ec'], 
					  results['runtime_generate_preimage'], results['runtime_total'],
					  results['itrs'], results['converged'],
					  results['num_updates_ecc'], 
					  results['mge']['num_decrease_order'] > 0, # @todo: not suitable for multi-start mge
					  results['mge']['num_increase_order'] > 0,
					  results['mge']['num_converged_descents'] > 0, 
					  nb_sod_sm2gm, 
					  nb_dis_k_sm2gm, nb_dis_k_gi2sm, nb_dis_k_gi2gm])
			f_summary.close()
			 
		# save median graphs.
		if save_medians:
			os.makedirs(dir_save + 'medians/', exist_ok=True)
			print('Saving median graphs to files...')
			fn_pre_sm = dir_save + 'medians/set_median.' + mpg_options['fit_method'] + '.nbg' + str(num_graphs) + '.y' + str(target) + '.repeat' + str(1)
			saveGXL(mpg.set_median, fn_pre_sm + '.gxl', method='default', 
					node_labels=dataset.node_labels, edge_labels=dataset.edge_labels,	
					node_attrs=dataset.node_attrs, edge_attrs=dataset.edge_attrs)
			fn_pre_gm = dir_save + 'medians/gen_median.' + mpg_options['fit_method'] + '.nbg' + str(num_graphs) + '.y' + str(target) + '.repeat' + str(1)
			saveGXL(mpg.gen_median, fn_pre_gm + '.gxl', method='default',
		  			node_labels=dataset.node_labels, edge_labels=dataset.edge_labels,	
					node_attrs=dataset.node_attrs, edge_attrs=dataset.edge_attrs)
			fn_best_dataset = dir_save + 'medians/g_best_dataset.' + mpg_options['fit_method'] + '.nbg' + str(num_graphs) + '.y' + str(target) + '.repeat' + str(1)
			saveGXL(mpg.best_from_dataset, fn_best_dataset + '.gxl', method='default',
		  			node_labels=dataset.node_labels, edge_labels=dataset.edge_labels,	
					node_attrs=dataset.node_attrs, edge_attrs=dataset.edge_attrs)
		
		# plot median graphs.
		if plot_medians and save_medians:
			if ged_options['edit_cost'] == 'LETTER2' or ged_options['edit_cost'] == 'LETTER' or ds_name == 'Letter-high' or ds_name == 'Letter-med' or ds_name == 'Letter-low':			
				draw_Letter_graph(mpg.set_median, fn_pre_sm)
				draw_Letter_graph(mpg.gen_median, fn_pre_gm)
				draw_Letter_graph(mpg.best_from_dataset, fn_best_dataset)
				
		if (load_gm == 'auto' and not gmfile_exist) or not load_gm:
			gram_matrix_unnorm_list.append(mpg.gram_matrix_unnorm)

	# write result summary for each class. 
	if save_results:
		sod_sm_mean = np.mean(sod_sm_list)
		sod_gm_mean = np.mean(sod_gm_list)
		dis_k_sm_mean = np.mean(dis_k_sm_list)
		dis_k_gm_mean = np.mean(dis_k_gm_list)
		dis_k_gi_min_mean = np.mean(dis_k_gi_min_list)
		time_precompute_gm_mean = np.mean(time_precompute_gm_list)
		time_optimize_ec_mean = np.mean(time_optimize_ec_list)
		time_generate_mean = np.mean(time_generate_list)
		time_total_mean = np.mean(time_total_list)
		itrs_mean = np.mean(itrs_list)
		num_converged = np.sum(converged_list)
		num_updates_ecc_mean = np.mean(num_updates_ecc_list)
		num_mge_decrease_order = np.sum(mge_decrease_order_list)
		num_mge_increase_order = np.sum(mge_increase_order_list)
		num_mge_converged = np.sum(mge_converged_order_list)
		sod_sm2gm_mean = get_relations(np.sign(sod_gm_mean - sod_sm_mean))
		dis_k_sm2gm_mean = get_relations(np.sign(dis_k_gm_mean - dis_k_sm_mean))
		dis_k_gi2sm_mean = get_relations(np.sign(dis_k_sm_mean - dis_k_gi_min_mean))
		dis_k_gi2gm_mean = get_relations(np.sign(dis_k_gm_mean - dis_k_gi_min_mean))
		f_summary = open(dir_save + fn_output_summary, 'a')
		csv.writer(f_summary).writerow([ds_name, kernel_options['name'], 
				  ged_options['edit_cost'], ged_options['method'], 
				  ged_options['attr_distance'], mpg_options['fit_method'], 
				  num_graphs, 'all',
				  sod_sm_mean, sod_gm_mean, dis_k_sm_mean, dis_k_gm_mean,
				  dis_k_gi_min_mean, sod_sm2gm_mean, dis_k_sm2gm_mean, 
				  dis_k_gi2sm_mean, dis_k_gi2gm_mean,
				  time_precompute_gm_mean, time_optimize_ec_mean,
				  time_generate_mean, time_total_mean, itrs_mean, 
				  num_converged, num_updates_ecc_mean,
				  num_mge_decrease_order, num_mge_increase_order,
				  num_mge_converged])
		f_summary.close()
		
	# save total pairwise kernel distances.
	dis_k_max = np.max(dis_k_max_list)
	dis_k_min = np.min(dis_k_min_list)
	dis_k_mean = np.mean(dis_k_mean_list)
	print('The maximum pairwise distance in kernel space:', dis_k_max)
	print('The minimum pairwise distance in kernel space:', dis_k_min)
	print('The average pairwise distance in kernel space:', dis_k_mean)
	
	# write Gram matrices to file.
	if (load_gm == 'auto' and not gmfile_exist) or not load_gm:
		np.savez(dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm', gram_matrix_unnorm_list=gram_matrix_unnorm_list, run_time_list=time_precompute_gm_list)	

	print('\ncomplete.\n')	

	
def _init_output_file_preimage(ds_name, gkernel, fit_method, dir_output):
	os.makedirs(dir_output, exist_ok=True)
#	fn_output_detail = 'results_detail.' + ds_name + '.' + gkernel + '.' + fit_method + '.csv'
	fn_output_detail = 'results_detail.' + ds_name + '.' + gkernel + '.csv'
	f_detail = open(dir_output + fn_output_detail, 'a')
	csv.writer(f_detail).writerow(['dataset', 'graph kernel', 'edit cost', 
			  'GED method', 'attr distance', 'fit method', 'num graphs', 
			  'target', 'repeat', 'SOD SM', 'SOD GM', 'dis_k SM', 'dis_k GM',
			  'min dis_k gi', 'SOD SM -> GM', 'dis_k SM -> GM', 'dis_k gi -> SM', 
			  'dis_k gi -> GM', 'edit cost constants', 'time precompute gm',
			  'time optimize ec', 'time generate preimage', 'time total',
			  'itrs', 'converged', 'num updates ecc', 'mge decrease order', 
			  'mge increase order', 'mge converged'])
	f_detail.close()
	
#	fn_output_summary = 'results_summary.' + ds_name + '.' + gkernel + '.' + fit_method + '.csv'
	fn_output_summary = 'results_summary.' + ds_name + '.' + gkernel + '.csv'
	f_summary = open(dir_output + fn_output_summary, 'a')
	csv.writer(f_summary).writerow(['dataset', 'graph kernel', 'edit cost', 
			  'GED method', 'attr distance', 'fit method', 'num graphs', 
			  'target', 'SOD SM', 'SOD GM', 'dis_k SM', 'dis_k GM',
			  'min dis_k gi', 'SOD SM -> GM', 'dis_k SM -> GM', 'dis_k gi -> SM', 
			  'dis_k gi -> GM', 'time precompute gm', 'time optimize ec', 
			  'time generate preimage', 'time total', 'itrs', 'num converged', 
			  'num updates ecc', 'mge num decrease order', 'mge num increase order', 
			  'mge num converged', '# SOD SM -> GM', '# dis_k SM -> GM', 
			  '# dis_k gi -> SM', '# dis_k gi -> GM'])
#			   'repeats better SOD SM -> GM', 
#			  'repeats better dis_k SM -> GM', 'repeats better dis_k gi -> SM', 
#			  'repeats better dis_k gi -> GM'])
	f_summary.close()
	
	return fn_output_detail, fn_output_summary


def get_relations(sign):
	if sign == -1:
		return 'better'
	elif sign == 0:
		return 'same'
	elif sign == 1:
		return 'worse'
	
	
#Dessin median courrant
def draw_Letter_graph(graph, file_prefix):
	import matplotlib
	matplotlib.use('agg')
	import matplotlib.pyplot as plt
	plt.figure()
	pos = {}
	for n in graph.nodes:
		pos[n] = np.array([float(graph.nodes[n]['x']),float(graph.nodes[n]['y'])])
	nx.draw_networkx(graph, pos)
	plt.savefig(file_prefix + '.eps', format='eps', dpi=300)
#	plt.show()
	plt.clf()
	plt.close()


def remove_edges(Gn):
	for G in Gn:
		for _, _, attrs in G.edges(data=True):
			attrs.clear()
			
			
def dis_gstar(idx_g, idx_gi, alpha, Kmatrix, term3=0, withterm3=True):
	term1 = Kmatrix[idx_g, idx_g]
	term2 = 0
	for i, a in enumerate(alpha):
		term2 += a * Kmatrix[idx_g, idx_gi[i]]
	term2 *= 2
	if withterm3 == False:
		for i1, a1 in enumerate(alpha):
			for i2, a2 in enumerate(alpha):
				term3 += a1 * a2 * Kmatrix[idx_gi[i1], idx_gi[i2]]
	return np.sqrt(term1 - term2 + term3)


def compute_k_dis(idx_g, idx_gi, alphas, Kmatrix, term3=0, withterm3=True):
	term1 = Kmatrix[idx_g, idx_g]
	term2 = 0
	for i, a in enumerate(alphas):
		term2 += a * Kmatrix[idx_g, idx_gi[i]]
	term2 *= 2
	if withterm3 == False:
		for i1, a1 in enumerate(alphas):
			for i2, a2 in enumerate(alphas):
				term3 += a1 * a2 * Kmatrix[idx_gi[i1], idx_gi[i2]]
	return np.sqrt(term1 - term2 + term3)


def compute_kernel(Gn, graph_kernel, node_label, edge_label, verbose, parallel='imap_unordered'):
	if graph_kernel == 'marginalizedkernel':
		Kmatrix, _ = marginalizedkernel(Gn, node_label=node_label, edge_label=edge_label,
								  p_quit=0.03, n_iteration=10, remove_totters=False,
								  n_jobs=multiprocessing.cpu_count(), verbose=verbose)
	elif graph_kernel == 'untilhpathkernel':
		Kmatrix, _ = untilhpathkernel(Gn, node_label=node_label, edge_label=edge_label,
								  depth=7, k_func='MinMax', compute_method='trie',
								  parallel=parallel,
								  n_jobs=multiprocessing.cpu_count(), verbose=verbose)
	elif graph_kernel == 'spkernel':
		mixkernel = functools.partial(kernelproduct, deltakernel, gaussiankernel)
		Kmatrix = np.empty((len(Gn), len(Gn)))
#		Kmatrix[:] = np.nan
		Kmatrix, _, idx = spkernel(Gn, node_label=node_label, node_kernels=
							  {'symb': deltakernel, 'nsymb': gaussiankernel, 'mix': mixkernel},
							  n_jobs=multiprocessing.cpu_count(), verbose=verbose)
#		for i, row in enumerate(idx):
#			for j, col in enumerate(idx):
#				Kmatrix[row, col] = Kmatrix_tmp[i, j]
	elif graph_kernel == 'structuralspkernel':
		mixkernel = functools.partial(kernelproduct, deltakernel, gaussiankernel)
		sub_kernels = {'symb': deltakernel, 'nsymb': gaussiankernel, 'mix': mixkernel}
		Kmatrix, _ = structuralspkernel(Gn, node_label=node_label, 
							  edge_label=edge_label, node_kernels=sub_kernels,
							  edge_kernels=sub_kernels,
							  parallel=parallel, n_jobs=multiprocessing.cpu_count(), 
							  verbose=verbose)
	elif graph_kernel == 'treeletkernel':
		pkernel = functools.partial(polynomialkernel, d=2, c=1e5)
#		pkernel = functools.partial(gaussiankernel, gamma=1e-6)
		mixkernel = functools.partial(kernelproduct, deltakernel, gaussiankernel)
		Kmatrix, _ = treeletkernel(Gn, node_label=node_label, edge_label=edge_label,
								   sub_kernel=pkernel, parallel=parallel,
								   n_jobs=multiprocessing.cpu_count(), verbose=verbose)
	elif graph_kernel == 'weisfeilerlehmankernel':
		Kmatrix, _ = weisfeilerlehmankernel(Gn, node_label=node_label, edge_label=edge_label,
								   height=4, base_kernel='subtree', parallel=None,
								   n_jobs=multiprocessing.cpu_count(), verbose=verbose)
	else:
		raise Exception('The graph kernel "', graph_kernel, '" is not defined.')	
		
	# normalization
	Kmatrix_diag = Kmatrix.diagonal().copy()
	for i in range(len(Kmatrix)):
		for j in range(i, len(Kmatrix)):
			Kmatrix[i][j] /= np.sqrt(Kmatrix_diag[i] * Kmatrix_diag[j])
			Kmatrix[j][i] = Kmatrix[i][j]
	return Kmatrix
			

def gram2distances(Kmatrix):
	dmatrix = np.zeros((len(Kmatrix), len(Kmatrix)))
	for i1 in range(len(Kmatrix)):
		for i2 in range(len(Kmatrix)):
			dmatrix[i1, i2] = Kmatrix[i1, i1] + Kmatrix[i2, i2] - 2 * Kmatrix[i1, i2]
	dmatrix = np.sqrt(dmatrix)
	return dmatrix


def kernel_distance_matrix(Gn, node_label, edge_label, Kmatrix=None, 
						   gkernel=None, verbose=True):
	import warnings
	warnings.warn('gklearn.preimage.utils.kernel_distance_matrix is deprecated, use gklearn.kernels.graph_kernel.compute_distance_matrix or gklearn.utils.compute_distance_matrix instead', DeprecationWarning)
	dis_mat = np.empty((len(Gn), len(Gn)))
	if Kmatrix is None:
		Kmatrix = compute_kernel(Gn, gkernel, node_label, edge_label, verbose)
	for i in range(len(Gn)):
		for j in range(i, len(Gn)):
			dis = Kmatrix[i, i] + Kmatrix[j, j] - 2 * Kmatrix[i, j]
			if dis < 0:
				if dis > -1e-10:
					dis = 0
				else:
					raise ValueError('The distance is negative.')
			dis_mat[i, j] = np.sqrt(dis)
			dis_mat[j, i] = dis_mat[i, j]
	dis_max = np.max(np.max(dis_mat))
	dis_min = np.min(np.min(dis_mat[dis_mat != 0]))
	dis_mean = np.mean(np.mean(dis_mat))
	return dis_mat, dis_max, dis_min, dis_mean


def get_same_item_indices(ls):
	"""Get the indices of the same items in a list. Return a dict keyed by items.
	"""
	idx_dict = {}
	for idx, item in enumerate(ls):
		if item in idx_dict:
			idx_dict[item].append(idx)
		else:
			idx_dict[item] = [idx]
	return idx_dict


def k_nearest_neighbors_to_median_in_kernel_space(Gn, Kmatrix=None, gkernel=None,
												  node_label=None, edge_label=None):
	dis_k_all = [] # distance between g_star and each graph.
	alpha = [1 / len(Gn)] * len(Gn)
	if Kmatrix is None:
		Kmatrix = compute_kernel(Gn, gkernel, node_label, edge_label, True)
	term3 = 0
	for i1, a1 in enumerate(alpha):
		for i2, a2 in enumerate(alpha):
			term3 += a1 * a2 * Kmatrix[idx_gi[i1], idx_gi[i2]]
	for ig, g in tqdm(enumerate(Gn_init), desc='computing distances', file=sys.stdout):
		dtemp = dis_gstar(ig, idx_gi, alpha, Kmatrix, term3=term3)
		dis_all.append(dtemp)


def normalize_distance_matrix(D):
	max_value = np.amax(D)
	min_value = np.amin(D)
	return (D - min_value) / (max_value - min_value)


# -*-coding:utf-8 -*-
"""
gklearn

This  package contains 4 sub  packages :
        * c_ext : binders to C++ code
        * ged : allows to compute graph edit distance between networkX graphs
        * kernels : computation of graph kernels, ie graph similarity measure compatible with SVM
        * notebooks : examples of code using this library
        * utils : Diverse computation on graphs
"""

# info
__version__ = "0.1"
__author__  = "Benoit Gaüzère"
__date__    = "November 2017"
 
# import sub modules
# from gklearn import c_ext
# from gklearn import ged
# from gklearn import utils



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  7 15:24:58 2020

@author: ljia

@references:
	[1] Borgwardt KM, Kriegel HP. Shortest-path kernels on graphs. InData
	Mining, Fifth IEEE International Conference on 2005 Nov 27 (pp. 8-pp). IEEE.
"""

import sys
from itertools import product
from multiprocessing import Pool
from gklearn.utils import get_iters
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError
import numpy as np
import networkx as nx
from gklearn.utils.parallel import parallel_gm, parallel_me
from gklearn.utils.utils import get_sp_graph
from gklearn.kernels import GraphKernel


class ShortestPath(GraphKernel):

	def __init__(self, **kwargs):
		GraphKernel.__init__(
			self, **{
				k: kwargs.get(k) for k in
				['parallel', 'n_jobs', 'chunksize', 'normalize', 'copy_graphs',
				 'verbose'] if k in kwargs}
		)
		self._node_labels = kwargs.get('node_labels', [])
		self._node_attrs = kwargs.get('node_attrs', [])
		self._edge_weight = kwargs.get('edge_weight', None)
		self._node_kernels = kwargs.get('node_kernels', None)
		self._fcsp = kwargs.get('fcsp', True)
		self._ds_infos = kwargs.get('ds_infos', {})
		self._save_sp_graphs = kwargs.get('save_sp_graphs', True)
		self._sp_func = get_sp_graph
		self._kernel_do_func = self._sp_do


	##########################################################################
	# The following is the 1st paradigm to compute kernel matrix, which is
	# compatible with `scikit-learn`.
	# -------------------------------------------------------------------
	# Special thanks to the "GraKeL" library for providing an excellent template!
	##########################################################################


	def clear_attributes(self):
		super().clear_attributes()
		if hasattr(self, '_sp_graphs'):
			delattr(self, '_sp_graphs')
		if hasattr(self, '_Y_sp_graphs'):
			delattr(self, '_Y_sp_graphs')


	def validate_parameters(self):
		super().validate_parameters()


	# if self._depth < 1:
	# 	raise ValueError('`depth` must be greater than 0.')
	# if self._k_func not in ['MinMax', 'tanimoto']:
	# 	raise ValueError('`k_func` must be either `MinMax` or `tanimoto`.')
	# if self._compute_method not in ['trie']:
	# 	raise ValueError('`compute_method` must be `trie`.')


	def _compute_kernel_matrix_series(self, Y, X=None, load_sp_graphs=True):
		"""Compute the kernel matrix between a given target graphs (Y) and
		the fitted graphs (X / self._graphs) without parallelization.

		Parameters
		----------
		Y : list of graphs, optional
			The target graphs.

		Returns
		-------
		kernel_matrix : numpy array, shape = [n_targets, n_inputs]
			The computed kernel matrix.

		"""
		if_comp_X_sp_graphs = True

		# if load saved sp_graphs of X from the instance:
		if load_sp_graphs:
			# sp_graphs for self._graphs.
			try:
				check_is_fitted(self, ['_sp_graphs'])
				sp_graphs_list1 = self._sp_graphs
				if_comp_X_sp_graphs = False
			except NotFittedError:
				import warnings
				warnings.warn(
					'The sp_graphs of self._graphs are not computed/saved. '
					'The sp_graphs of `X` is computed instead.'
				)
				if_comp_X_sp_graphs = True

		# Get all sp_graphs of all graphs before computing kernels to save
		# time, but this may cost a lot of memory for large dataset.

		# Compute the sp_graphs of X.
		if if_comp_X_sp_graphs:
			if X is None:
				raise ('X can not be None.')
			# self._add_dummy_labels will modify the input in place.
			self._all_graphs_have_edges(X)  # for X
			sp_graphs_list1 = []
			iterator = get_iters(
				self._graphs, desc='Getting sp_graphs for X',
				file=sys.stdout, verbose=(self.verbose >= 2)
			)
			for g in iterator:
				sp_graphs_list1.append(self._sp_func(g))

		# sp_graphs for Y.
		# 		Y = [g.copy() for g in Y] # @todo: ?
		self._all_graphs_have_edges(Y)
		sp_graphs_list2 = []
		iterator = get_iters(
			Y, desc='Getting sp_graphs for Y', file=sys.stdout,
			verbose=(self.verbose >= 2)
		)
		for g in iterator:
			sp_graphs_list2.append(self._sp_func(g))

		# 		if self.save_sp_graphs:
		# 			self._Y_sp_graphs = sp_graphs_list2

		# compute kernel matrix.
		kernel_matrix = np.zeros((len(Y), len(sp_graphs_list1)))

		from itertools import product
		itr = product(range(len(Y)), range(len(sp_graphs_list1)))
		len_itr = int(len(Y) * len(sp_graphs_list1))
		iterator = get_iters(
			itr, desc='Computing kernels', file=sys.stdout,
			length=len_itr, verbose=(self.verbose >= 2)
		)
		for i_y, i_x in iterator:
			kernel = self._kernel_do_func(
				sp_graphs_list2[i_y], sp_graphs_list1[i_x]
			)
			kernel_matrix[i_y][i_x] = kernel

		return kernel_matrix


	def _compute_kernel_matrix_imap_unordered(self, Y):
		"""Compute the kernel matrix between a given target graphs (Y) and
		the fitted graphs (X / self._graphs) using imap unordered parallelization.

		Parameters
		----------
		Y : list of graphs, optional
			The target graphs.

		Returns
		-------
		kernel_matrix : numpy array, shape = [n_targets, n_inputs]
			The computed kernel matrix.

		"""
		raise NotImplementedError(
			'Parallelization for kernel matrix is not implemented.'
		)


	def pairwise_kernel(self, x, y, are_sp_graphs=False):
		"""Compute pairwise kernel between two graphs.

		Parameters
		----------
		x, y : NetworkX Graph.
			Graphs bewteen which the kernel is computed.

		are_sp_graphs : boolean, optional
			If `True`, `x` and `y` are sp graphs, otherwise are graphs.
			The default is False.

		Returns
		-------
		kernel: float
			The computed kernel.

		"""
		if are_sp_graphs:
			# x, y are canonical keys.
			kernel = self._kernel_do_func(x, y)

		else:
			# x, y are graphs.
			kernel = self._compute_single_kernel_series(x, y)

		return kernel


	def diagonals(self):
		"""Compute the kernel matrix diagonals of the fit/transformed data.

		Returns
		-------
		X_diag : numpy array
			The diagonal of the kernel matrix between the fitted data.
			This consists of each element calculated with itself.

		Y_diag : numpy array
			The diagonal of the kernel matrix, of the transform.
			This consists of each element calculated with itself.
		"""
		# Check if method "fit" had been called.
		check_is_fitted(self, ['_graphs'])

		# Check if the diagonals of X exist.
		try:
			check_is_fitted(self, ['_X_diag'])
		except NotFittedError:
			# Compute diagonals of X.
			self._X_diag = np.empty(shape=(len(self._graphs),))
			try:
				check_is_fitted(self, ['_all_sp_graphs'])
				for i, x in enumerate(self._all_graphs):
					self._X_diag[i] = self.pairwise_kernel(
						x, x, are_sp_graphs=True
					)  # @todo: parallel?
			except NotFittedError:
				for i, x in enumerate(self._graphs):
					self._X_diag[i] = self.pairwise_kernel(
						x, x, are_sp_graphs=False
					)  # @todo: parallel?

		try:
			# If transform has happened, return both diagonals.
			check_is_fitted(self, ['_Y'])
			self._Y_diag = np.empty(shape=(len(self._Y),))
			try:
				check_is_fitted(self, ['_Y_all_sp_graphs'])
				for (i, y) in enumerate(self._Y_all_sp_graphs):
					self._Y_diag[i] = self.pairwise_kernel(
						y, y, are_sp_graphs=True
					)  # @todo: parallel?
			except NotFittedError:
				for (i, y) in enumerate(self._Y):
					self._Y_diag[i] = self.pairwise_kernel(
						y, y, are_sp_graphs=False
					)  # @todo: parallel?

			return self._X_diag, self._Y_diag

		except NotFittedError:
			# Else just return both X_diag
			return self._X_diag


	##########################################################################
	# The following is the 2nd paradigm to compute kernel matrix. It is
	# simplified and not compatible with `scikit-learn`.
	##########################################################################


	def _compute_gm_series(self, graphs):
		self._all_graphs_have_edges(graphs)
		# get shortest path graph of each graph.
		iterator = get_iters(
			graphs, desc='getting sp graphs', file=sys.stdout,
			verbose=(self.verbose >= 2)
		)
		graphs = [
			get_sp_graph(g, edge_weight=self._edge_weight) for g in
			iterator
		]
		if self._save_sp_graphs:
			self._sp_graphs = graphs

		# compute Gram matrix.
		gram_matrix = np.zeros((len(graphs), len(graphs)))

		from itertools import combinations_with_replacement
		itr = combinations_with_replacement(range(0, len(graphs)), 2)
		len_itr = int(len(graphs) * (len(graphs) + 1) / 2)
		iterator = get_iters(
			itr, desc='Computing kernels',
			length=len_itr, file=sys.stdout, verbose=(self.verbose >= 2)
		)
		for i, j in iterator:
			kernel = self._sp_do(graphs[i], graphs[j])
			gram_matrix[i][j] = kernel
			gram_matrix[j][i] = kernel

		return gram_matrix


	def _compute_gm_imap_unordered(self):
		self._all_graphs_have_edges(self._graphs)
		# get shortest path graph of each graph.
		pool = Pool(self.n_jobs)
		get_sp_graphs_fun = self._wrapper_get_sp_graphs
		itr = zip(self._graphs, range(0, len(self._graphs)))
		if len(self._graphs) < 100 * self.n_jobs:
			chunksize = int(len(self._graphs) / self.n_jobs) + 1
		else:
			chunksize = 100
		iterator = get_iters(
			pool.imap_unordered(get_sp_graphs_fun, itr, chunksize),
			desc='getting sp graphs', file=sys.stdout,
			length=len(self._graphs), verbose=(self.verbose >= 2)
		)
		for i, g in iterator:
			self._graphs[i] = g
		pool.close()
		pool.join()

		if self._save_sp_graphs:
			self._sp_graphs = self._graphs  # @TODO: deepcopy? save original graphs?

		# compute Gram matrix.
		gram_matrix = np.zeros((len(self._graphs), len(self._graphs)))


		def init_worker(gs_toshare):
			global G_gs
			G_gs = gs_toshare


		do_fun = self._wrapper_sp_do
		parallel_gm(
			do_fun, gram_matrix, self._graphs, init_worker=init_worker,
			glbv=(self._graphs,), n_jobs=self.n_jobs, verbose=self.verbose
		)

		return gram_matrix


	def _compute_kernel_list_series(self, g1, g_list):
		self._all_graphs_have_edges([g1] + g_list)
		# get shortest path graphs of g1 and each graph in g_list.
		g1 = get_sp_graph(g1, edge_weight=self._edge_weight)
		iterator = get_iters(
			g_list, desc='getting sp graphs', file=sys.stdout,
			verbose=(self.verbose >= 2)
		)
		g_list = [
			get_sp_graph(g, edge_weight=self._edge_weight) for g in
			iterator
		]

		# compute kernel list.
		kernel_list = [None] * len(g_list)
		iterator = get_iters(
			range(len(g_list)), desc='Computing kernels', file=sys.stdout,
			length=len(g_list), verbose=(self.verbose >= 2)
		)
		for i in iterator:
			kernel = self._sp_do(g1, g_list[i])
			kernel_list[i] = kernel

		return kernel_list


	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		self._all_graphs_have_edges([g1] + g_list)
		# get shortest path graphs of g1 and each graph in g_list.
		g1 = get_sp_graph(g1, edge_weight=self._edge_weight)
		pool = Pool(self.n_jobs)
		get_sp_graphs_fun = self._wrapper_get_sp_graphs
		itr = zip(g_list, range(0, len(g_list)))
		if len(g_list) < 100 * self.n_jobs:
			chunksize = int(len(g_list) / self.n_jobs) + 1
		else:
			chunksize = 100
		iterator = get_iters(
			pool.imap_unordered(get_sp_graphs_fun, itr, chunksize),
			desc='getting sp graphs', file=sys.stdout,
			length=len(g_list), verbose=(self.verbose >= 2)
		)
		for i, g in iterator:
			g_list[i] = g
		pool.close()
		pool.join()

		# compute Gram matrix.
		kernel_list = [None] * len(g_list)


		def init_worker(g1_toshare, gl_toshare):
			global G_g1, G_gl
			G_g1 = g1_toshare
			G_gl = gl_toshare


		do_fun = self._wrapper_kernel_list_do


		def func_assign(result, var_to_assign):
			var_to_assign[result[0]] = result[1]


		itr = range(len(g_list))
		len_itr = len(g_list)
		parallel_me(
			do_fun, func_assign, kernel_list, itr, len_itr=len_itr,
			init_worker=init_worker, glbv=(g1, g_list), method='imap_unordered',
			n_jobs=self.n_jobs, itr_desc='Computing kernels',
			verbose=self.verbose
		)

		return kernel_list


	def _wrapper_kernel_list_do(self, itr):
		return itr, self._sp_do(G_g1, G_gl[itr])


	def _compute_single_kernel_series(self, g1, g2):
		self._all_graphs_have_edges([g1] + [g2])
		g1 = get_sp_graph(g1, edge_weight=self._edge_weight)
		g2 = get_sp_graph(g2, edge_weight=self._edge_weight)
		kernel = self._sp_do(g1, g2)
		return kernel


	def _wrapper_get_sp_graphs(self, itr_item):
		g = itr_item[0]
		i = itr_item[1]
		return i, get_sp_graph(g, edge_weight=self._edge_weight)


	def _sp_do(self, g1, g2):

		if self._fcsp:  # @todo: it may be put outside the _sp_do().
			return self._sp_do_fcsp(g1, g2)
		else:
			return self._sp_do_naive(g1, g2)


	def _sp_do_fcsp(self, g1, g2):

		kernel = 0

		# compute shortest path matrices first, method borrowed from FCSP.
		vk_dict = {}  # shortest path matrices dict
		if len(
				self._node_labels
		) > 0:  # @todo: it may be put outside the _sp_do().
			# node symb and non-synb labeled
			if len(self._node_attrs) > 0:
				kn = self._node_kernels['mix']
				for n1, n2 in product(
						g1.nodes(data=True), g2.nodes(data=True)
				):
					n1_labels = [n1[1][nl] for nl in self._node_labels]
					n2_labels = [n2[1][nl] for nl in self._node_labels]
					# @TODO: reformat attrs during data processing a priori to save time.
					n1_attrs = np.array(
						[n1[1][na] for na in self._node_attrs]
					).astype(float)
					n2_attrs = np.array(
						[n2[1][na] for na in self._node_attrs]
					).astype(float)
					vk_dict[(n1[0], n2[0])] = kn(
						n1_labels, n2_labels, n1_attrs, n2_attrs
					)
			# node symb labeled
			else:
				kn = self._node_kernels['symb']
				for n1 in g1.nodes(data=True):
					for n2 in g2.nodes(data=True):
						n1_labels = [n1[1][nl] for nl in self._node_labels]
						n2_labels = [n2[1][nl] for nl in self._node_labels]
						vk_dict[(n1[0], n2[0])] = kn(n1_labels, n2_labels)
		else:
			# node non-synb labeled
			if len(self._node_attrs) > 0:
				kn = self._node_kernels['nsymb']
				for n1 in g1.nodes(data=True):
					for n2 in g2.nodes(data=True):
						n1_attrs = np.array(
							[n1[1][na] for na in self._node_attrs]
						).astype(float)
						n2_attrs = np.array(
							[n2[1][na] for na in self._node_attrs]
						).astype(float)
						vk_dict[(n1[0], n2[0])] = kn(n1_attrs, n2_attrs)
			# node unlabeled
			else:
				for e1, e2 in product(
						g1.edges(data=True), g2.edges(data=True)
				):
					if e1[2]['cost'] == e2[2]['cost']:
						kernel += 1
				return kernel

		# compute graph kernels
		if self._ds_infos['directed']:
			for e1, e2 in product(g1.edges(data=True), g2.edges(data=True)):
				if e1[2]['cost'] == e2[2]['cost']:
					nk11, nk22 = vk_dict[(e1[0], e2[0])], vk_dict[
						(e1[1], e2[1])]
					kn1 = nk11 * nk22
					kernel += kn1
		else:
			for e1, e2 in product(g1.edges(data=True), g2.edges(data=True)):
				if e1[2]['cost'] == e2[2]['cost']:
					# each edge walk is counted twice, starting from both its extreme nodes.
					nk11, nk12, nk21, nk22 = vk_dict[(e1[0], e2[0])], vk_dict[(
						e1[0], e2[1])], vk_dict[(e1[1], e2[0])], vk_dict[
						(e1[1], e2[1])]
					kn1 = nk11 * nk22
					kn2 = nk12 * nk21
					kernel += kn1 + kn2

		# # ---- exact implementation of the Fast Computation of Shortest Path Kernel (FCSP), reference [2], sadly it is slower than the current implementation
		# # compute vertex kernels
		# try:
		#	 vk_mat = np.zeros((nx.number_of_nodes(g1),
		#						nx.number_of_nodes(g2)))
		#	 g1nl = enumerate(g1.nodes(data=True))
		#	 g2nl = enumerate(g2.nodes(data=True))
		#	 for i1, n1 in g1nl:
		#		 for i2, n2 in g2nl:
		#			 vk_mat[i1][i2] = kn(
		#				 n1[1][node_label], n2[1][node_label],
		#				 [n1[1]['attributes']], [n2[1]['attributes']])

		#	 range1 = range(0, len(edge_w_g[i]))
		#	 range2 = range(0, len(edge_w_g[j]))
		#	 for i1 in range1:
		#		 x1 = edge_x_g[i][i1]
		#		 y1 = edge_y_g[i][i1]
		#		 w1 = edge_w_g[i][i1]
		#		 for i2 in range2:
		#			 x2 = edge_x_g[j][i2]
		#			 y2 = edge_y_g[j][i2]
		#			 w2 = edge_w_g[j][i2]
		#			 ke = (w1 == w2)
		#			 if ke > 0:
		#				 kn1 = vk_mat[x1][x2] * vk_mat[y1][y2]
		#				 kn2 = vk_mat[x1][y2] * vk_mat[y1][x2]
		#				 kernel += kn1 + kn2

		return kernel


	def _sp_do_naive(self, g1, g2):

		kernel = 0

		# Define the function to compute kernels between vertices in each condition.
		if len(self._node_labels) > 0:
			# node symb and non-synb labeled
			if len(self._node_attrs) > 0:
				def compute_vk(n1, n2):
					kn = self._node_kernels['mix']
					n1_labels = [g1.nodes[n1][nl] for nl in self._node_labels]
					n2_labels = [g2.nodes[n2][nl] for nl in self._node_labels]
					# @TODO: reformat attrs during data processing a priori to save time.
					n1_attrs = np.array(
						[g1.nodes[n1][na] for na in self._node_attrs]
					).astype(float)
					n2_attrs = np.array(
						[g2.nodes[n2][na] for na in self._node_attrs]
					).astype(float)
					return kn(n1_labels, n2_labels, n1_attrs, n2_attrs)
			# node symb labeled
			else:
				def compute_vk(n1, n2):
					kn = self._node_kernels['symb']
					n1_labels = [g1.nodes[n1][nl] for nl in self._node_labels]
					n2_labels = [g2.nodes[n2][nl] for nl in self._node_labels]
					return kn(n1_labels, n2_labels)
		else:
			# node non-synb labeled
			if len(self._node_attrs) > 0:
				def compute_vk(n1, n2):
					kn = self._node_kernels['nsymb']
					n1_attrs = np.array(
						[g1.nodes[n1][na] for na in self._node_attrs]
					).astype(float)
					n2_attrs = np.array(
						[g2.nodes[n2][na] for na in self._node_attrs]
					).astype(float)
					return kn(n1_attrs, n2_attrs)
			# node unlabeled
			else:
				for e1, e2 in product(g1.edges(data=True), g2.edges(data=True)):
					if e1[2]['cost'] == e2[2]['cost']:
						kernel += 1
				return kernel

		# compute graph kernels
		if self._ds_infos['directed']:
			for e1, e2 in product(g1.edges(data=True), g2.edges(data=True)):
				if e1[2]['cost'] == e2[2]['cost']:
					nk11, nk22 = compute_vk(e1[0], e2[0]), compute_vk(
						e1[1], e2[1]
					)
					kn1 = nk11 * nk22
					kernel += kn1
		else:
			for e1, e2 in product(g1.edges(data=True), g2.edges(data=True)):
				if e1[2]['cost'] == e2[2]['cost']:
					# each edge walk is counted twice, starting from both its extreme nodes.
					nk11, nk12, nk21, nk22 = compute_vk(
						e1[0], e2[0]
					), compute_vk(
						e1[0], e2[1]
					), compute_vk(e1[1], e2[0]), compute_vk(e1[1], e2[1])
					kn1 = nk11 * nk22
					kn2 = nk12 * nk21
					kernel += kn1 + kn2

		return kernel


	def _wrapper_sp_do(self, itr):
		i = itr[0]
		j = itr[1]
		return i, j, self._sp_do(G_gs[i], G_gs[j])


	@staticmethod
	def _all_graphs_have_edges(graphs):
		for G in graphs:
			if nx.number_of_edges(G) == 0:
				raise ValueError('Not all graphs have edges!!!')
